import base64
import re


def file_to_base64(path):
    with open(path, 'rb') as f:
        content = f.read()
    return base64.b64encode(content).decode()


def check_schema(url: str) -> str:
    url = url.strip('/')
    if not re.findall(r'(http://|https://)', url):
        return "http://" + url
    return url



import functools

from botoy.action import Action
from botoy.model import GroupMsg

# import time


@functools.lru_cache(520)
def __get_group_admins(qq, host, port, group):
    admins = Action(qq=qq, port=port, host=host).getGroupAdminList(group, True)
    return [admin['MemberUin'] for admin in admins]


def from_admin(func=None):
    """来自群管理员(列表包括群主)的消息 GroupMsg
    管理员列表会进行``缓存``，调用520次后再次刷新, 所以可以放心使用"""
    if func is None:
        return from_admin

    def inner(ctx):
        assert isinstance(ctx, GroupMsg)
        # start = time.time()
        admins = __get_group_admins(
            qq=ctx.CurrentQQ,
            # pylint: disable=W0212
            host=ctx._host,
            port=ctx._port,
            group=ctx.FromGroupId,
        )
        # print('===================')
        # print('coast {}'.format(time.time() - start))
        # print('===================')
        if ctx.FromUserId in admins:
            return func(ctx)
        return None

    return inner



import collections
import threading
import time
import traceback
from typing import List, Union

import httpx

from botoy import macro
from botoy.config import Config
from botoy.log import logger


class Action:
    def __init__(self, qq: int, port: int = None, host: str = None, timeout=20):
        self.qq = qq
        self.config = Config(host=host, port=port)
        self.c = httpx.Client(
            headers={'Content-Type': 'application/json'},
            timeout=timeout + 5,
            base_url=self.config.address,
            params={'qq': self.qq, 'timeout': timeout},
        )
        self.lock = threading.Lock()

    ############发送相关############
    def sendFriendText(self, user: int, content: str) -> dict:
        """发送好友文本消息"""
        return self.post(
            'SendMsgV2',
            {
                "ToUserUid": user,
                "SendToType": 1,
                "SendMsgType": "TextMsg",
                "Content": content,
            },
        )

    def sendFriendPic(
        self,
        user: int,
        *,
        picUrl: str = '',
        picBase64Buf: str = '',
        fileMd5: str = '',
        content: str = '',
        flashPic=False,
    ):
        """发送好友图片消息"""
        assert any([picUrl, picBase64Buf, fileMd5]), '缺少参数'
        return self.post(
            'SendMsg',
            {
                "toUser": user,
                "sendToType": 1,
                "sendMsgType": "PicMsg",
                "content": content,
                "picUrl": picUrl,
                "picBase64Buf": picBase64Buf,
                "fileMd5": fileMd5,
                "flashPic": flashPic,
            },
        )

    def sendFriendVoice(
        self, user: int, *, voiceUrl: str = '', voiceBase64Buf: str = ''
    ):
        """发送好友语音消息"""
        assert any([voiceUrl, voiceBase64Buf]), '缺少参数'
        return self.post(
            'SendMsg',
            {
                "toUser": user,
                "sendToType": 1,
                "sendMsgType": "VoiceMsg",
                "voiceUrl": voiceUrl,
                "voiceBase64Buf": voiceBase64Buf,
            },
        )

    def sendFriendXml(self, user: int, content: str) -> dict:
        """发送好友Xml消息"""
        return self.post(
            'SendMsgV2',
            {
                "ToUserUid": user,
                "SendToType": 1,
                "SendMsgType": "XmlMsg",
                "Content": content,
            },
        )

    def sendFriendTeXiaoText(self, user: int, text: str) -> dict:
        """发送好友特效文本消息"""
        return self.post(
            'SendMsgV2',
            {
                "ToUserUid": user,
                "SendToType": 1,
                "SendMsgType": "TeXiaoTextMsg",
                "Content": text,
            },
        )

    def sendGroupText(
        self, group: int, content: str, atUser: Union[int, List[int]] = 0
    ) -> dict:
        """发送群组文本消息"""
        if atUser != 0:
            content = macro.atUser(atUser) + '\n' + content
        return self.post(
            'SendMsgV2',
            {
                "ToUserUid": group,
                "SendToType": 2,
                "SendMsgType": "TextMsg",
                "Content": content,
            },
        )

    def sendGroupPic(
        self,
        group: int,
        *,
        content: str = '',
        picUrl: str = '',
        picBase64Buf: str = '',
        fileMd5: str = '',
        picMd5s: Union[str, List[str]] = '',
        flashPic=False,
        atUser: Union[int, List[int]] = 0,
    ) -> dict:
        """发送群组图片消息"""
        assert any([picUrl, picBase64Buf, fileMd5, picMd5s]), '缺少参数'
        if atUser != 0:
            content = macro.atUser(atUser) + '\n' + content
        if picMd5s:
            if not isinstance(picMd5s, collections.Sequence):
                picMd5s = [picMd5s]
            return self.post(
                'SendMsgV2',
                {
                    "ToUserUid": group,
                    "SendToType": 2,
                    "SendMsgType": "PicMsg",
                    "PicMd5s": picMd5s,
                },
            )
        return self.post(
            'SendMsg',
            {
                "toUser": group,
                "sendToType": 2,
                "sendMsgType": "PicMsg",
                "content": content,
                "picUrl": picUrl,
                "picBase64Buf": picBase64Buf,
                "fileMd5": fileMd5,
                "flashPic": flashPic,
            },
        )

    def sendGroupVoice(
        self, group: int, *, voiceUrl: str = '', voiceBase64Buf: str = ''
    ) -> dict:
        """发送群组语音消息"""
        assert any([voiceUrl, voiceBase64Buf]), '缺少参数'
        return self.post(
            'SendMsg',
            {
                "toUser": group,
                "sendToType": 2,
                "sendMsgType": "VoiceMsg",
                "voiceUrl": voiceUrl,
                "voiceBase64Buf": voiceBase64Buf,
            },
        )

    def sendGroupXml(self, group: int, content: str) -> dict:
        """发送群组Xml消息"""
        return self.post(
            'SendMsgV2',
            {
                "ToUserUid": group,
                "SendToType": 2,
                "SendMsgType": "XmlMsg",
                "Content": content,
            },
        )

    def sendGroupJson(self, group: int, content: str) -> dict:
        """发送群组Json消息"""
        return self.post(
            'SendMsgV2',
            {
                "ToUserUid": group,
                "SendToType": 2,
                "SendMsgType": "JsonMsg",
                "Content": content,
            },
        )

    def sendGroupTeXiaoText(self, group: int, text: str) -> dict:
        """发送群组特效文本消息"""
        return self.post(
            'SendMsgV2',
            {
                "ToUserUid": group,
                "SendToType": 2,
                "SendMsgType": "TeXiaoTextMsg",
                "Content": text,
            },
        )

    def sendPrivateText(self, user: int, group: int, content: str) -> dict:
        """发送私聊文本消息"""
        return self.post(
            'SendMsgV2',
            {
                "ToUserUid": user,
                "GroupID": group,
                "SendToType": 3,
                "SendMsgType": "TextMsg",
                "Content": content,
            },
        )

    def sendPrivateXml(self, user: int, group: int, content: str) -> dict:
        """发送私聊Xml消息"""
        return self.post(
            'SendMsgV2',
            {
                "ToUserUid": user,
                "GroupID": group,
                "SendToType": 3,
                "SendMsgType": "XmlMsg",
                "Content": content,
            },
        )

    def sendPrivateVoice(
        self, user: int, group: int, *, voiceUrl: str = '', voiceBase64Buf: str = ''
    ) -> dict:
        assert any([voiceUrl, voiceBase64Buf]), '缺少参数'
        return self.post(
            'SendMsg',
            {
                "toUser": user,
                "sendToType": 3,
                "sendMsgType": "VoiceMsg",
                "content": "",
                "groupid": group,
                "voiceUrl": voiceUrl,
                "voiceBase64Buf": voiceBase64Buf,
            },
        )

    def sendPrivatePic(
        self,
        user: int,
        group: int,
        content: str = '',
        *,
        picUrl: str = '',
        picBase64Buf: str = '',
        fileMd5: str = '',
    ) -> dict:
        """发送私聊图片消息"""
        assert any([picUrl, picBase64Buf, fileMd5]), '缺少参数'
        return self.post(
            'SendMsg',
            {
                "toUser": user,
                "sendToType": 3,
                "sendMsgType": "PicMsg",
                "content": content,
                "groupid": group,
                "picUrl": picUrl,
                "picBase64Buf": picBase64Buf,
                "fileMd5": fileMd5,
            },
        )

    def sendPhoneText(self, content: str):
        """给手机发文字"""
        return self.post(
            'SendMsgV2',
            {"SendToType": 2, "SendMsgType": "PhoneMsg", "Content": content},
        )

    def replyGroupMsg(
        self,
        group: int,
        content: str,
        msgSeq: int,
        msgTime: int = None,
        user: int = 0,
        rawContent: str = '',
    ):
        """发送回复消息, 回复群消息
        下面的原消息表示需要回复的消息
        :param group: 原消息的群号
        :param content: 回复内容
        :param msgSeq: 原消息的msgSeq, 点击跳转到该条消息位置
        :param msgTime: 原消息的msgTime, 如果不指定，默认为当前时间戳
        :param user: 原消息的人的qq号，也可以是其他人，该用户收到消息会提示“有新回复”, 默认为0
        :param rawContent: 原消息内容，可以任意指定，默认为空
        """
        return self.post(
            "SendMsg",
            {
                "toUser": group,
                "sendToType": 2,
                "sendMsgType": "ReplayMsg",
                "content": content,
                "replayInfo": {
                    "MsgSeq": msgSeq,
                    "MsgTime": msgTime or int(time.time()),
                    "UserID": user,
                    "RawContent": rawContent,
                },
            },
        )

    def replyFriendMsg(
        self,
        user: int,
        content: str,
        msgSeq: int,
        msgTime: int = None,
        rawContent: str = '',
    ):
        """发送回复消息, 回复好友消息
        下面的原消息表示需要回复的消息
        :param user: 原消息发送人
        :param content: 回复内容
        :param msgSeq: 原消息的msgSeq, 点击跳转到该条消息位置
        :param msgTime: 原消息的msgTime, 如果不指定，默认为当前时间戳
        :param rawContent: 原消息内容，可以任意指定，默认为空
        """
        return self.post(
            "SendMsg",
            {
                "toUser": user,
                "sendToType": 1,
                "sendMsgType": "ReplayMsg",
                "content": content,
                "replayInfo": {
                    "MsgSeq": msgSeq,
                    "MsgTime": msgTime or int(time.time()),
                    "UserID": user,
                    "RawContent": rawContent,
                },
            },
        )

    def repostVideo2Group(self, group: int, forwordBuf: str) -> dict:
        """转发视频到群聊
        :param group: 群号
        :param forwordBuf: 原视频的forwordBuf字段
        """
        return self.post(
            'SendMsg',
            {
                "toUser": group,
                "sendToType": 2,
                "sendMsgType": "ForwordMsg",
                "forwordBuf": forwordBuf,
                "forwordField": 19,
            },
        )

    def getVideoURL(self, group: int, videoMD5: str, videoURL: str) -> dict:
        """获取视频链接
        :param group: 群号
        :param videoMD5: 视频的 MD5 值，可以从上报的消息中获得
        :param videoURL: 上报的消息中传给的 url 值
        """
        return self.post(
            'PttCenterSvr.ShortVideoDownReq',
            {"GroupID": group, "VideoUrl": videoURL, "VideoMd5": videoMD5},
        )

    def getFriendFileURL(self, fileID: str) -> dict:
        """获取好友文件链接
        :param fileID: 文件 ID，可由上报的消息中获得
        """
        return self.post(
            'OfflineFilleHandleSvr.pb_ftn_CMD_REQ_APPLY_DOWNLOAD-1200',
            {"FileID": fileID},
        )

    def getGroupFileURL(self, group: int, fileID: str) -> dict:
        """获取好友文件链接
        :param group: 群号
        :param fileID: 文件 ID，可由上报的消息中获得
        """
        return self.post(
            'OidbSvc.0x6d6_2',
            {'GroupID': group, "FileID": fileID},
        )

    def repostVideo2Friend(self, user: int, forwordBuf: str) -> dict:
        """转发视频到好友
        :param user: 好友QQ
        :param forwordBuf: 原视频的forwordBuf字段
        """
        return self.post(
            'SendMsg',
            {
                "toUser": user,
                "sendToType": 1,
                "sendMsgType": "ForwordMsg",
                "forwordBuf": forwordBuf,
                "forwordField": 19,
            },
        )

    ############获取############
    def getCookies(self) -> dict:
        """获取QQ相关cookie"""
        return self.get('GetUserCook')

    def getUserInfo(self, user: int) -> dict:
        """获取用户信息昵称头像等"""
        return self.post('GetUserInfo', {'UserID': user})

    def getSummaryCard(self, user: int) -> dict:
        """获取企鹅卡片资料"""
        return self.post('SummaryCard.ReqSummaryCard', {'UserID': user})

    def getUserList(self) -> List[dict]:
        """获取好友列表"""
        friend_list = []
        start_index = 0
        while True:
            data = self.post('GetQQUserList', {'StartIndex': start_index})
            if 'Friendlist' not in data:
                break
            friend_list.extend(data['Friendlist'])
            if len(friend_list) >= int(data.get('Totoal_friend_count', 0)):  # 这里有个拼写错误
                break
            if "GetfriendCount" not in data:
                break
            start_index += int(
                data.get("GetfriendCount", 999999)
            )  # 设置 999999 为了在 API 坏了的情况下能跳出循环
        return friend_list

    def getGroupList(self) -> List[dict]:
        """获取群列表"""
        next_token = ""
        group_list = []
        while True:
            data = self.post('GetGroupList', {'NextToken': next_token})
            if 'TroopList' not in data:
                break
            group_list.extend(data['TroopList'])
            next_token = data.get('NextToken', '')
            if not next_token:
                break
        return group_list

    def getGroupMembers(self, group: int) -> List[dict]:
        """获取群成员列表"""
        members = []
        lastUin = 0
        while True:
            data = self.post(
                'GetGroupUserList', {"GroupUin": group, "LastUin": lastUin}
            )
            if 'MemberList' in data:
                members.extend(data['MemberList'])
            if 'LastUin' not in data or data['LastUin'] == 0:
                break
            lastUin = data['LastUin']
            time.sleep(0.6)
        return members

    def getGroupAdminList(self, group: int, include_owner=True) -> List[dict]:
        """获取群管理员列表
        :param group: 群号
        :param include_owner: 是否包括群主
        """
        members = self.getGroupMembers(group)
        if include_owner:
            # 获取群主id
            for groupInfo in self.getGroupList():
                if groupInfo['GroupId'] == group:
                    owner = groupInfo['GroupOwner']
                    break
            admins = [
                member
                for member in members
                if member['GroupAdmin'] == 1 or member['MemberUin'] == owner
            ]
        else:
            admins = [member for member in members if member['GroupAdmin'] == 1]
        return admins

    def getClusterInfo(self) -> dict:
        """获取当前集群信息"""
        return self.get('', path='/v1/ClusterInfo')

    ############操作############
    def setUniqueTitle(self, user: int, group: int, title: str) -> dict:
        """设置群成员头衔"""
        return self.post(
            'OidbSvc.0x8fc_2',
            {"GroupID": group, "UserID": user, "NewTitle": title},
        )

    def modifyGroupCard(self, user: int, group: int, nick: str) -> dict:
        """修改群名片"""
        return self.post(
            'ModifyGroupCard', {'UserID': user, 'GroupID': group, 'NewNick': nick}
        )

    def shutUserUp(self, groupID: int, userid: int, ShutTime: int) -> dict:
        """禁言用户(禁言时间单位为分钟 ShutTime=0 取消禁言)"""
        return self.post(
            'ShutUp',
            {
                'ShutUpType': 0,
                'GroupID': groupID,
                'ShutUid': userid,
                'ShutTime': ShutTime,
            },
        )

    def shutAllUp(self, group: int, switch: int = 1):
        """全体禁言
        :param switch: 1 开启; 0 关闭, 默认为1即开启全体禁言
        """
        return self.post(
            'OidbSvc.0x89a_0',
            {"GroupID": group, "Switch": 0 if switch == 0 else 1},
        )

    def setGroupAnnounce(
        self,
        group: int,
        text: str,
        pinned: bool = False,
        title: str = '',
        typ: bool = True,
    ):
        """设置群公告
        :param group: 群号
        :param text: 内容
        :param pinned: 是否置顶
        :param title: 标题,可以空
        :param typ: 是否发送新成员
        """
        return self.post(
            path='/v1/Group/Announce',
            funcname='',
            payload={
                "GroupID": group,
                "Title": title,
                "Text": text,
                "Pinned": 1 if pinned else 0,
                "Type": 20 if typ else 10,  # 20 发给新成员, 10 弹窗
            },
        )

    def likeUser(self, user: int, cmd=0) -> dict:
        """给某人点赞
        :param user: 用户QQ号
        :param cmd: 发送包选项, 0 或 1; 0表示使用``QQZan``; 1表示使用``OidbSvc.0x7e5_4``, 默认为0
        """
        return self.post('QQZan' if cmd == 0 else 'OidbSvc.0x7e5_4', {"UserID": user})

    def toggleGroupAdmin(self, user: int, group: int, flag=1) -> dict:
        """设置和取消群管理员
        :param user: 用户QQ
        :param group: 群号
        :param flag: 1 或 0, 1表示设置为管理员; 0 表示取消管理员, 默认为1
        """
        return self.post(
            'OidbSvc.0x55c_1',
            {"GroupID": group, "UserID": user, "Flag": 0 if flag == 0 else 1},
        )

    def revokeGroupMsg(
        self, group: int, msgSeq: int, msgRandom: int, cmd: int = 0
    ) -> dict:
        """撤回群消息
        :param group: 群号
        :param msgSeq: 消息msgSeq
        :param msgRandom: 消息msgRandom
        :param cmd: 0 或 1, 0表示使用: RevokeMsg , 1表示使用: PbMessageSvc.PbMsgWithDraw
        """
        return self.post(
            'RevokeMsg' if cmd == 0 else 'PbMessageSvc.PbMsgWithDraw',
            {"GroupID": group, "MsgSeq": msgSeq, "MsgRandom": msgRandom},
        )

    def inviteUserJoinGroup(self, group: int, user: int) -> dict:
        """拉人入群
        :param group: 哪个群?
        :param user: 拉谁?
        """
        return self.post(
            'GroupMgr',
            {"ActionType": 8, "GroupID": group, "ActionUserID": user, "Content": ""},
        )

    def joinGroup(self, group: int, content: str = "") -> dict:
        """加入群聊
        :param group: 哪个群?
        :param content: 加群理由
        """
        return self.post(
            'GroupMgr',
            {"ActionType": 1, "GroupID": group, "ActionUserID": 0, "Content": content},
        )

    def exitGroup(self, group: int) -> dict:
        """退出群聊
        :param group: 哪个群?
        """
        return self.post(
            'GroupMgr',
            {"ActionType": 2, "GroupID": group, "ActionUserID": 0, "Content": ""},
        )

    def driveUserAway(self, group: int, user: int) -> dict:
        """移出群聊
        :param group: 哪个群?
        :param user:把谁踢出去?
        """
        return self.post(
            'GroupMgr',
            {"ActionType": 3, "GroupID": group, "ActionUserID": user, "Content": ""},
        )

    def refreshKeys(self) -> dict:
        """刷新key二次登陆"""
        return self.get('', path='/v1/RefreshKeys')

    def logout(self, flag=False) -> bool:
        """退出指定QQ
        :param flag: 是否删除设备信息文件
        """
        return self.post('LogOut', {"Flag": flag})

    ############################################################################
    def baseRequest(
        self,
        method: str,
        funcname: str,
        path: str,
        payload: dict = None,
        params: dict = None,
    ) -> dict:
        """基础请求方法, 提供部分提示信息，出错返回空字典，其他返回服务端响应结果"""
        if params is not None:
            params = params.update({'funcname': funcname})
        else:
            params = {'funcname': funcname}

        # 发送请求
        try:
            self.lock.acquire()
            threading.Timer(1, self.release_lock).start()
            resp = self.c.request(
                method, httpx.URL(url=path, params=params), json=payload
            )
            resp.raise_for_status()
        except httpx.HTTPError as e:
            if isinstance(e, httpx.TimeoutException):
                logger.warning(f'响应超时，但不代表处理未成功, 结果未知! => {e}')
            elif isinstance(e, httpx.HTTPStatusError):
                logger.error(
                    f"响应码出错 => {resp.status_code}",
                )
            else:
                logger.error(f'请求出错: {traceback.format_exc()}')
            return {}

        # 处理数据
        try:
            data = resp.json()
        except Exception as e:
            logger.error('API响应结果非json格式')
            return {}

        if data is None:
            logger.error('返回为null')
            return {}

        # 返回码提示
        if 'Ret' in data:
            ret = data.get('Ret')
            if ret == 0:
                pass
            elif ret == 34:
                logger.error(f'未知错误，跟消息长度似乎无关，可以尝试分段重新发送 => {data}')
            elif ret == 110:
                logger.error(f'发送失败，你已被移出该群，请重新加群 => {data}')
            elif ret == 120:
                logger.error(f'机器人被禁言 => {data}')
            elif ret == 241:
                logger.error(f'消息发送频率过高，对同一个群或好友，建议发消息的最小间隔控制在1100ms以上 => {data}')
            elif ret == 299:
                logger.error(f'超过群发言频率限制 => {data}')
            else:
                logger.error(f'请求发送成功, 但处理失败 => {data}')

        return data

    def post(
        self,
        funcname: str,
        payload: dict,
        params: dict = None,
        path: str = '/v1/LuaApiCaller',
    ) -> dict:
        """封装常用的post操作"""
        return self.baseRequest(
            'POST', funcname=funcname, path=path, payload=payload, params=params
        )

    def get(
        self,
        funcname: str,
        params: dict = None,
        path: str = '/v1/LuaApiCaller',
    ) -> dict:
        """封装get操作"""
        return self.baseRequest('GET', funcname=funcname, path=path, params=params)

    def release_lock(self):
        self.lock.release()



from ..model import FriendMsg, GroupMsg


def ignore_these_users(*users):
    """忽略这些人的消息 GroupMsg, FriendMsg"""

    def deco(func):
        def inner(ctx):
            nonlocal users
            assert isinstance(ctx, (GroupMsg, FriendMsg))
            if isinstance(ctx, GroupMsg):
                from_user = ctx.FromUserId
            elif isinstance(ctx, FriendMsg):
                from_user = ctx.FromUin
            if from_user not in users:
                return func(ctx)
            return None

        return inner

    return deco



from ..model import FriendMsg


def ignore_tempMsg(func=None):
    """忽略私聊信息 FriendMsg"""
    if func is None:
        return ignore_tempMsg

    def inner(ctx):
        assert isinstance(ctx, FriendMsg)
        if ctx.TempUin is None:
            return func(ctx)
        return None

    return inner



# pylint: disable=C0413
"""解析各类消息的详细字段, 所有操作不会影响原始消息上下文对象"""
import copy
import functools


def _copy_ctx(f):
    @functools.wraps(f)
    def i(ctx):
        return f(copy.deepcopy(ctx))

    return i


from ._event import (
    _FriendDeleteEventMsg,
    _FriendRevokeEventMsg,
    _GroupAdminEventMsg,
    _GroupAdminsysnotifyEventMsg,
    _GroupExitEventMsg,
    _GroupJoinEventMsg,
    _GroupRevokeEventMsg,
    _GroupShutEventMsg,
    refine_friend_delete_event_msg,
    refine_friend_revoke_event_msg,
    refine_group_admin_event_msg,
    refine_group_adminsysnotify_event_msg,
    refine_group_exit_event_msg,
    refine_group_join_event_msg,
    refine_group_revoke_event_msg,
    refine_group_shut_event_msg
)
from ._friend_msg import (
    _FriendFileMsg,
    _FriendPic,
    _PicFriendMsg,
    _RedBagFriendMsg,
    _ReplyFriendMsg,
    _VideoFriendMsg,
    _VoiceFriendMsg,
    refine_file_friend_msg,
    refine_pic_friend_msg,
    refine_RedBag_friend_msg,
    refine_reply_friend_msg,
    refine_video_friend_msg,
    refine_voice_friend_msg
)
from ._group_msg import (
    _AtGroupMsg,
    _GroupFileMsg,
    _GroupPic,
    _PicGroupMsg,
    _RedBagGroupMsg,
    _ReplyGroupMsg,
    _VideoGroupMsg,
    _VoiceGroupMsg,
    refine_at_group_msg,
    refine_file_group_msg,
    refine_pic_group_msg,
    refine_RedBag_group_msg,
    refine_reply_group_msg,
    refine_video_group_msg,
    refine_voice_group_msg
)



from ..collection import MsgTypes
from ..model import FriendMsg


def from_phone(func=None):
    """来自手机的消息(给自己发的) FriendMsg"""
    if func is None:
        return from_phone

    def inner(ctx):
        assert isinstance(ctx, FriendMsg)
        if ctx.MsgType == MsgTypes.PhoneMsg:
            return func(ctx)
        return None

    return inner



from ..model import FriendMsg, GroupMsg
from ..refine import refine_pic_friend_msg, refine_pic_group_msg


def equal_content(string: str):
    """发送的内容与指定字符串相等时 GroupMsg, FriendMsg"""

    def deco(func):
        def inner(ctx):
            assert isinstance(ctx, (GroupMsg, FriendMsg))
            if isinstance(ctx, GroupMsg):
                pic_ctx = refine_pic_group_msg(ctx)
            else:
                pic_ctx = refine_pic_friend_msg(ctx)
            if pic_ctx is not None:
                content = pic_ctx.Content
            else:
                content = ctx.Content
            if content == string:
                return func(ctx)
            return None

        return inner

    return deco



import re

from ..model import FriendMsg, GroupMsg
from ..refine import refine_pic_friend_msg, refine_pic_group_msg


def in_content(string: str, raw: bool = True):
    """Content字段包括指定字符串  GroupMsg, FriendMsg

    :param string: 需要包含的字符串, 支持使用正则查找
    :param raw: 为True则使用最原始的Content字段数据进行查找, 即图片这类消息会包含图片链接、
                图片MD5之类的数据; 为False则会提取真正发送的文字部分内容
    """

    def deco(func):
        def inner(ctx):
            assert isinstance(ctx, (GroupMsg, FriendMsg))
            if raw:
                if re.match(string, ctx.Content):
                    return func(ctx)
            else:
                if isinstance(ctx, GroupMsg):
                    pic_ctx = refine_pic_group_msg(ctx)
                else:
                    pic_ctx = refine_pic_friend_msg(ctx)
                if pic_ctx is not None:
                    content = pic_ctx.Content
                else:
                    content = ctx.Content
                if re.match(string, content):
                    return func(ctx)
            return None

        return inner

    return deco



import re

from ..model import FriendMsg, GroupMsg


def with_pattern(pattern: str):
    """正则匹配Content字段 GroupMsg, FriendMsg
    因为使用这种功能一般匹配的内容都比较特殊,像图片，视频之类的消息基本是不会符合匹配条件的,
    所以不会解析特殊的消息, 均采用最原始的Content字段进行匹配,

    注意: 成功匹配才会执行该接收函数,匹配到的结果是一个列表,该结果会增加为ctx的``_pattern_result``属性

    :param pattern: 需要进行匹配的``正则表达式字符串``
    """

    def deco(func):
        def inner(ctx):
            assert isinstance(ctx, (GroupMsg, FriendMsg))
            result = re.findall(pattern, ctx.Content)
            if result:
                ctx._pattern_result = result  # pylint: disable=W0212
                return func(ctx)
            return None

        return inner

    return deco



import traceback
from concurrent.futures import ThreadPoolExecutor

from botoy.log import logger


class WorkerPool(ThreadPoolExecutor):
    def resize(self, workers: int):
        self._max_workers = workers

    def callback(self, worker):
        worker_exception = worker.exception()
        if worker_exception:
            try:
                raise worker_exception
            except Exception:
                logger.error(traceback.format_exc())

    def submit(self, *args, **kwargs):
        super().submit(*args, **kwargs).add_done_callback(self.callback)



# pylint: disable=R0902,W0231
from typing import Dict, List

from botoy import json
from botoy.collection import MsgTypes
from botoy.exceptions import InvalidContextError
from botoy.model import FriendMsg
from botoy.refine import _copy_ctx


class _FriendMsg(FriendMsg):
    def _carry_properties(self, ctx: FriendMsg):
        self.message = ctx.message
        self.CurrentQQ = ctx.CurrentQQ

        self.data = ctx.data

        self.FromUin: int = ctx.FromUin
        self.ToUin: int = ctx.ToUin
        self.MsgType: str = ctx.MsgType
        self.MsgSeq: int = ctx.MsgSeq
        self.Content: str = ctx.Content
        self.RedBaginfo: dict = ctx.RedBaginfo


class _VoiceFriendMsg(_FriendMsg):
    """好友语音消息"""

    def __init__(self, ctx: FriendMsg):
        voice_data = json.loads(ctx.Content)
        self.VoiceUrl: str = voice_data['Url']
        self.Tips: str = voice_data['Tips']
        super()._carry_properties(ctx)


class _VideoFriendMsg(_FriendMsg):
    """好友视频消息"""

    def __init__(self, ctx: FriendMsg):
        video_data = json.loads(ctx.Content)
        self.ForwordBuf: str = video_data['ForwordBuf']
        self.ForwordField: int = video_data['ForwordField']
        self.Tips: str = video_data['Tips']
        self.VideoMd5: str = video_data['VideoMd5']
        self.VideoSize: str = video_data['VideoSize']
        self.VideoUrl: str = video_data['VideoUrl']
        super()._carry_properties(ctx)


class _FriendPic:
    def __init__(self, pic: dict):
        """好友图片单个图片所包含的数据
        [{"FileMd5":"","FileSize":0,"Path":"","Url":""}]中的一个
        """
        self.FileMd5: str = pic.get('FileMd5')
        self.FileSize: int = pic.get('FileSize')
        self.Path: str = pic.get('Path')
        self.Url: str = pic.get('Url')


class _PicFriendMsg(_FriendMsg):
    """好友图片/表情包消息"""

    def __init__(self, ctx: FriendMsg):
        pic_data = json.loads(ctx.Content)
        self.FriendPic: List[_FriendPic] = [
            _FriendPic(i) for i in pic_data['FriendPic']
        ]
        self.Tips: str = pic_data['Tips']
        super()._carry_properties(ctx)
        self.Content = pic_data.get('Content') or ''


class _RedBagFriendMsg(_FriendMsg):
    """好友红包消息"""

    def __init__(self, ctx: FriendMsg):
        redbag_info = ctx.RedBaginfo
        self.RedBag_Authkey: str = redbag_info.get('Authkey')
        self.RedBag_Channel: int = redbag_info.get('Channel')
        self.RedBag_Des: str = redbag_info.get('Des')
        self.RedBag_FromType: int = redbag_info.get('FromType')
        self.RedBag_FromUin: int = redbag_info.get('FromUin')
        self.RedBag_Listid: str = redbag_info.get('Listid')
        self.RedBag_RedType: int = redbag_info.get('RedType')
        self.RedBag_StingIndex: str = redbag_info.get('StingIndex')
        self.RedBag_Tittle: str = redbag_info.get('Tittle')
        self.RedBag_Token_17_2: str = redbag_info.get('Token_17_2')
        self.RedBag_Token_17_3: str = redbag_info.get('Token_17_3')
        super()._carry_properties(ctx)


class _ReplyFriendMsg(_FriendMsg):
    """好友语音消息"""

    def __init__(self, ctx: FriendMsg):
        reply_data: Dict = json.loads(ctx.Content)
        super()._carry_properties(ctx)
        self.Content: str = reply_data.get('Content', '')
        self.OriginMsgSeq: int = reply_data.get('MsgSeq', -1)
        self.SrcContent: str = reply_data.get('SrcContent', '')
        self.Tips: str = reply_data.get('Tips', '')
        self.AtUserID: List = reply_data.get('UserID', [])


class _FriendFileMsg(_FriendMsg):
    """好友文件消息"""

    def __init__(self, ctx: FriendMsg):
        file_data: Dict = json.loads(ctx.Content)
        super()._carry_properties(ctx)
        self.FileID: str = file_data.get('FileID', '')
        self.FileName: str = file_data.get('FileName', '')
        self.FileSize: int = file_data.get('FileSize', -1)
        self.Tips: str = file_data.get('Tips', '')


@_copy_ctx
def refine_voice_friend_msg(ctx: FriendMsg) -> _VoiceFriendMsg:
    """好友语音消息"""
    if not isinstance(ctx, FriendMsg):
        raise InvalidContextError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.VoiceMsg:
        return _VoiceFriendMsg(ctx)
    return None


@_copy_ctx
def refine_video_friend_msg(ctx: FriendMsg) -> _VideoFriendMsg:
    """好友视频消息"""
    if not isinstance(ctx, FriendMsg):
        raise InvalidContextError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.VideoMsg:
        return _VideoFriendMsg(ctx)
    return None


@_copy_ctx
def refine_pic_friend_msg(ctx: FriendMsg) -> _PicFriendMsg:
    """好友图片/表情包消息"""
    if not isinstance(ctx, FriendMsg):
        raise InvalidContextError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.PicMsg:
        return _PicFriendMsg(ctx)
    return None


@_copy_ctx
def refine_RedBag_friend_msg(ctx: FriendMsg) -> _RedBagFriendMsg:
    """好友红包消息"""
    if not isinstance(ctx, FriendMsg):
        raise InvalidContextError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.RedBagMsg:
        return _RedBagFriendMsg(ctx)
    return None


@_copy_ctx
def refine_reply_friend_msg(ctx: FriendMsg) -> _ReplyFriendMsg:
    """好友回复消息"""
    if not isinstance(ctx, FriendMsg):
        raise InvalidContextError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType in (MsgTypes.ReplyMsg, MsgTypes.ReplyMsgA):
        return _ReplyFriendMsg(ctx)
    return None


@_copy_ctx
def refine_file_friend_msg(ctx: FriendMsg) -> _FriendFileMsg:
    """好友文件消息"""
    if not isinstance(ctx, FriendMsg):
        raise InvalidContextError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.FriendFileMsg:
        return _FriendFileMsg(ctx)
    return None



# pylint: disable=R0902,W0231
from botoy.collection import EventNames
from botoy.exceptions import InvalidContextError
from botoy.model import EventMsg
from botoy.refine import _copy_ctx


class _EventMsg(EventMsg):
    def _carry_properties(self, ctx: EventMsg):
        self.message = ctx.message
        self.CurrentQQ = ctx.CurrentQQ

        self.data = ctx.data

        self.EventName = ctx.EventName
        self.EventData = ctx.EventData
        self.EventMsg = ctx.EventMsg

        self.Content = ctx.Content
        self.FromUin = ctx.FromUin
        self.MsgSeq = ctx.MsgSeq
        self.MsgType = ctx.MsgType
        self.ToUin = ctx.ToUin
        self.RedBaginfo = ctx.RedBaginfo


class _GroupRevokeEventMsg(_EventMsg):
    """群成员撤回消息事件"""

    def __init__(self, ctx: EventMsg):
        event_data = ctx.EventData
        self.AdminUserID: int = event_data.get('AdminUserID')
        self.GroupID: int = event_data.get('GroupID')
        self.MsgRandom: int = event_data.get('MsgRandom')
        self.MsgSeq: int = event_data.get('MsgSeq')
        self.UserID: int = event_data.get('UserID')
        super()._carry_properties(ctx)


class _GroupExitEventMsg(_EventMsg):
    """群成员退出群聊事件"""

    def __init__(self, ctx: EventMsg):
        self.UserID = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


class _GroupJoinEventMsg(_EventMsg):
    """某人进群事件"""

    def __init__(self, ctx: EventMsg):
        e_data = ctx.EventData
        self.InviteUin: int = e_data.get('InviteUin')
        self.UserID: int = e_data.get('UserID')
        self.UserName: str = e_data.get('UserName')
        super()._carry_properties(ctx)


class _FriendRevokeEventMsg(_EventMsg):
    """好友撤回消息事件"""

    def __init__(self, ctx: EventMsg):
        self.MsgSeq = ctx.EventData.get('MsgSeq')
        self.UserID = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


class _FriendDeleteEventMsg(_EventMsg):
    """删除好友事件"""

    def __init__(self, ctx: EventMsg):
        self.UserID: int = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


class _GroupAdminsysnotifyEventMsg(_EventMsg):
    """QQ群系统消息通知(加群申请在这里面"""

    def __init__(self, ctx: EventMsg):
        edata = ctx.EventData
        self.Type: int = edata.get('Type')  # 事件类型
        self.MsgTypeStr: str = edata.get('MsgTypeStr')  # 消息类型
        self.MsgStatusStr: str = edata.get('MsgStatusStr')  # 消息类型状态
        self.Who: int = edata.get('Who')  # 触发消息的对象
        self.WhoName: int = edata.get('WhoName')  # 触发消息的对象昵称
        self.GroupID: int = edata.get('GroupId')  # 来自群
        self.GroupName: str = edata.get('GroupName')  # 群名
        self.ActionUin: int = edata.get('ActionUin')  # 邀请人(处理人)
        self.ActionName: str = edata.get('ActionName')  # 邀请人(处理人)昵称
        self.ActionGroupCard: str = edata.get('ActionGroupCard')  # 邀请人(处理人)群名片
        self.Action: str = edata.get('Action')  # 加群理由 11 agree 14 忽略 12/21 disagree
        self.Content: int = edata.get('Content')
        super()._carry_properties(ctx)


class _GroupShutEventMsg(_EventMsg):
    """群禁言事件"""

    def __init__(self, ctx: EventMsg):
        self.GroupID: int = ctx.EventData.get('GroupID')
        self.ShutTime: int = ctx.EventData.get('ShutTime')
        self.UserID: int = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


class _GroupAdminEventMsg(_EventMsg):
    """管理员变更事件"""

    def __init__(self, ctx: EventMsg):
        self.Flag: int = ctx.EventData.get('Flag')
        self.GroupID: int = ctx.EventData.get('GroupID')
        self.UserID: int = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


@_copy_ctx
def refine_group_revoke_event_msg(ctx: EventMsg) -> _GroupRevokeEventMsg:
    """群成员撤回消息事件"""
    if not isinstance(ctx, EventMsg):
        raise InvalidContextError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_REVOKE:
        return _GroupRevokeEventMsg(ctx)
    return None


@_copy_ctx
def refine_group_exit_event_msg(ctx: EventMsg) -> _GroupExitEventMsg:
    """群成员退出群聊事件"""
    if not isinstance(ctx, EventMsg):
        raise InvalidContextError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_EXIT:
        return _GroupExitEventMsg(ctx)
    return None


@_copy_ctx
def refine_group_join_event_msg(ctx: EventMsg) -> _GroupJoinEventMsg:
    """某人进群事件"""
    if not isinstance(ctx, EventMsg):
        raise InvalidContextError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_JOIN:
        return _GroupJoinEventMsg(ctx)
    return None


@_copy_ctx
def refine_friend_revoke_event_msg(ctx: EventMsg) -> _FriendRevokeEventMsg:
    """好友撤回消息事件"""
    if not isinstance(ctx, EventMsg):
        raise InvalidContextError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_FRIEND_REVOKE:
        return _FriendRevokeEventMsg(ctx)
    return None


@_copy_ctx
def refine_friend_delete_event_msg(ctx: EventMsg) -> _FriendDeleteEventMsg:
    """删除好友事件"""
    if not isinstance(ctx, EventMsg):
        raise InvalidContextError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_FRIEND_DELETE:
        return _FriendDeleteEventMsg(ctx)
    return None


@_copy_ctx
def refine_group_adminsysnotify_event_msg(
    ctx: EventMsg,
) -> _GroupAdminsysnotifyEventMsg:
    """加群申请"""
    if not isinstance(ctx, EventMsg):
        raise InvalidContextError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_ADMINSYSNOTIFY:
        return _GroupAdminsysnotifyEventMsg(ctx)
    return None


@_copy_ctx
def refine_group_shut_event_msg(ctx: EventMsg) -> _GroupShutEventMsg:
    """群禁言事件"""
    if not isinstance(ctx, EventMsg):
        raise InvalidContextError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_SHUT:
        return _GroupShutEventMsg(ctx)
    return None


@_copy_ctx
def refine_group_admin_event_msg(ctx: EventMsg) -> _GroupAdminEventMsg:
    """管理员变更事件"""
    if not isinstance(ctx, EventMsg):
        raise InvalidContextError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_ADMIN:
        return _GroupAdminEventMsg(ctx)
    return None



"""See https://github.com/xiyaowong/botoy
"""
try:
    import ujson as json
except ImportError:
    import json

from . import decorators, refine
from .action import Action
from .async_action import AsyncAction
from .async_client import AsyncBotoy
from .client import Botoy
from .model import EventMsg, FriendMsg, GroupMsg

__all__ = [
    'refine',
    'decorators',
    'sugar',
    'collection',
    'Action',
    'AsyncAction',
    'Botoy',
    'AsyncBotoy',
    'EventMsg',
    'GroupMsg',
    'FriendMsg',
]

from .__version__ import check_version as __check_version

__check_version()



import sys

from loguru import logger

logger.remove()

logger.add(
    sys.stdout,
    format='{level.icon} {time:YYYY-MM-DD HH:mm:ss} <lvl>{level}\t{message}</lvl>',
    colorize=True,
)


def enble_log_file():
    logger.add(
        './logs/{time}.log',
        format='{time:YYYY-MM-DD HH:mm} {level}\t{message}',
        rotation='1 day',
        encoding='utf-8',
    )



class BotoyError(BaseException):
    pass


# class ThreadPoolError(BotoyError):
#     """线程池任务错误"""


class InvalidContextError(BotoyError):
    """不是正确的消息上下文对象"""


class InvalidConfigError(BotoyError):
    """配置文件有毛病"""


class InvalidPluginError(BotoyError):
    """插件问题"""



# pylint: disable=too-many-instance-attributes
class GroupMsg:
    def __init__(self, message: dict):
        self.message: dict = message
        self.CurrentQQ: int = message.get('CurrentQQ')

        temp = message.get('CurrentPacket')
        self.data: dict = temp.get('Data') if temp is not None else {}

        self.FromGroupId: int = self.data.get('FromGroupId')
        self.FromGroupName: str = self.data.get('FromGroupName')
        self.FromUserId: int = self.data.get('FromUserId')
        self.FromNickName: str = self.data.get('FromNickName')
        self.Content: str = self.data.get('Content')
        self.MsgType: str = self.data.get('MsgType')
        self.MsgTime: int = self.data.get('MsgTime')
        self.MsgSeq: int = self.data.get('MsgSeq')
        self.MsgRandom: int = self.data.get('MsgRandom')
        self.RedBaginfo: dict = self.data.get('RedBaginfo')

    def __getitem__(self, key):
        return self.message[key]


class FriendMsg:
    def __init__(self, message: dict):
        self.message: dict = message
        self.CurrentQQ: int = message.get('CurrentQQ')

        temp = message.get('CurrentPacket')
        self.data: dict = temp.get('Data') if temp is not None else {}

        self.FromUin: int = self.data.get('FromUin')
        self.ToUin: int = self.data.get('ToUin')
        self.MsgType: str = self.data.get('MsgType')
        self.MsgSeq: int = self.data.get('MsgSeq')
        self.Content: str = self.data.get('Content')
        self.RedBaginfo: dict = self.data.get('RedBaginfo')

        # 私聊(临时会话)特有
        self.TempUin: int = self.data.get('TempUin')  # 入口群聊ID

    def __getitem__(self, key):
        return self.message[key]


class EventMsg:
    def __init__(self, message: dict):
        self.message: dict = message
        self.CurrentQQ: int = message.get('CurrentQQ')

        temp = message.get('CurrentPacket')
        self.data: dict = temp.get('Data') if temp is not None else {}

        self.EventName: str = self.data.get('EventName')
        self.EventData: dict = self.data.get('EventData')
        self.EventMsg: dict = self.data.get('EventMsg')

        self.Content: str = self.EventMsg.get('Content')
        self.FromUin: int = self.EventMsg.get('FromUin')
        self.MsgSeq: int = self.EventMsg.get('MsgSeq')
        self.MsgType: str = self.EventMsg.get('MsgType')
        self.ToUin: int = self.EventMsg.get('ToUin')
        self.RedBaginfo = self.EventMsg.get('RedBaginfo')

    def __getitem__(self, key):
        return self.message[key]


model_map = {'OnGroupMsgs': GroupMsg, 'OnFriendMsgs': FriendMsg, 'OnEvents': EventMsg}



from ..model import FriendMsg, GroupMsg


def these_msgtypes(*msgtypes):
    """仅接受该些类型消息  GroupMsg, FriendMsg
    模块collection中定义了这些消息类型
    """

    def deco(func):
        def inner(ctx):
            assert isinstance(ctx, (GroupMsg, FriendMsg))
            if ctx.MsgType in msgtypes:
                return func(ctx)
            return None

        return inner

    return deco



# pylint: disable=W0212
"""深度封装Action方法
函数只能在群消息和好友消息接收函数中调用
"""
import sys

from .action import Action
from .collection import MsgTypes
from .exceptions import InvalidContextError
from .model import FriendMsg, GroupMsg
from .util import file_to_base64


def Text(text: str, at=False):
    """发送文字
    :param text: 文字内容
    :param at:是否艾特发送该消息的用户
    """
    text = str(text)
    # 查找消息上下文 `ctx`变量
    ctx = None
    f = sys._getframe()
    upper = f.f_back
    upper_locals = upper.f_locals
    if 'ctx' in upper_locals and isinstance(upper_locals['ctx'], (FriendMsg, GroupMsg)):
        ctx = upper_locals['ctx']
    else:
        for v in upper_locals.values():
            if isinstance(v, (GroupMsg, FriendMsg)):
                ctx = v
                break
    if ctx is None:
        raise InvalidContextError('经支持群消息和好友消息接收函数内调用')

    if hasattr(ctx, '_host') and hasattr(ctx, '_port'):
        action = Action(ctx.CurrentQQ, port=ctx._port, host=ctx._host)
    else:
        action = Action(ctx.CurrentQQ)

    if isinstance(ctx, GroupMsg):
        return action.sendGroupText(
            ctx.FromGroupId, text, atUser=ctx.FromUserId if at else 0
        )
    if isinstance(ctx, FriendMsg):
        if ctx.TempUin:  # 私聊消息
            return action.sendPrivateText(ctx.FromUin, ctx.TempUin, text)
        elif ctx.MsgType == MsgTypes.PhoneMsg:  # 来自手机的消息
            return action.sendPhoneText(text)
        else:
            return action.sendFriendText(ctx.FromUin, text)
    return None


def Picture(pic_url='', pic_base64='', pic_path='', pic_md5='', text=''):
    """发送图片 经支持群消息和好友消息接收函数内调用
    :param pic_url: 图片链接
    :param pic_base64: 图片base64编码
    :param pic_path: 图片文件路径
    :param pic_md5: 已发送图片的MD5
    :param text: 包含的文字消息

    ``pic_url, pic_base64, pic_path必须给定一项``
    """
    assert any([pic_url, pic_base64, pic_path, pic_md5]), '必须给定一项'

    ctx = None
    f = sys._getframe()
    upper = f.f_back
    upper_locals = upper.f_locals
    if 'ctx' in upper_locals and isinstance(upper_locals['ctx'], (FriendMsg, GroupMsg)):
        ctx = upper_locals['ctx']
    else:
        for v in upper_locals.values():
            if isinstance(v, (FriendMsg, GroupMsg)):
                ctx = v
                break
    if ctx is None:
        raise InvalidContextError('经支持群消息和好友消息接收函数内调用')

    if hasattr(ctx, '_host') and hasattr(ctx, '_port'):
        action = Action(ctx.CurrentQQ, port=ctx._port, host=ctx._host)
    else:
        action = Action(ctx.CurrentQQ)

    if isinstance(ctx, GroupMsg):
        if pic_url:
            return action.sendGroupPic(ctx.FromGroupId, content=text, picUrl=pic_url)

        elif pic_base64:
            return action.sendGroupPic(
                ctx.FromGroupId, content=text, picBase64Buf=pic_base64
            )
        elif pic_path:
            return action.sendGroupPic(
                ctx.FromGroupId, content=text, picBase64Buf=file_to_base64(pic_path)
            )
        elif pic_md5:
            return action.sendGroupPic(ctx.FromGroupId, content=text, fileMd5=pic_md5)

    if isinstance(ctx, FriendMsg):
        if pic_url:
            if ctx.TempUin is not None:
                return action.sendPrivatePic(
                    ctx.FromUin, ctx.TempUin, content=text, picUrl=pic_url
                )
            else:
                return action.sendFriendPic(ctx.FromUin, picUrl=pic_url, content=text)
        elif pic_base64:
            if ctx.TempUin:
                return action.sendPrivatePic(
                    ctx.FromUin, ctx.TempUin, content=text, picBase64Buf=pic_base64
                )
            elif ctx.MsgType == MsgTypes.PhoneMsg:  # 来自手机的消息
                return None
            else:
                return action.sendFriendPic(
                    ctx.FromUin, picBase64Buf=pic_base64, content=text
                )
        elif pic_path:
            if ctx.TempUin:
                return action.sendPrivatePic(
                    ctx.FromUin,
                    ctx.TempUin,
                    content=text,
                    picBase64Buf=file_to_base64(pic_path),
                )
            elif ctx.MsgType == MsgTypes.PhoneMsg:  # 来自手机的消息
                return None
            else:
                return action.sendFriendPic(
                    ctx.FromUin, picBase64Buf=file_to_base64(pic_path), content=text
                )
        elif pic_md5:
            if ctx.TempUin:
                return action.sendPrivatePic(
                    ctx.FromUin, ctx.TempUin, content=text, fileMd5=pic_md5
                )
            elif ctx.MsgType == MsgTypes.PhoneMsg:  # 来自手机的消息
                return None
            else:
                return action.sendFriendPic(ctx.FromUin, fileMd5=pic_md5, content=text)
    return None


def Voice(voice_url='', voice_base64='', voice_path=''):
    """发送语音 经支持群消息和好友消息接收函数内调用
    :param voice_url: 语音链接
    :param voice_base64: 语音base64编码
    :param voice_path: 语音文件路径
    voice_url, voice_base64, voice_path必须给定一项
    """
    assert any([voice_url, voice_base64, voice_path]), '必须给定一项'

    ctx = None
    f = sys._getframe()
    upper = f.f_back
    upper_locals = upper.f_locals
    if 'ctx' in upper_locals and isinstance(upper_locals['ctx'], (FriendMsg, GroupMsg)):
        ctx = upper_locals['ctx']
    else:
        for v in upper_locals.values():
            if isinstance(v, (GroupMsg, FriendMsg)):
                ctx = v
                break
    if ctx is None:
        raise InvalidContextError('经支持群消息和好友消息接收函数内调用')

    if hasattr(ctx, '_host') and hasattr(ctx, '_port'):
        action = Action(ctx.CurrentQQ, port=ctx._port, host=ctx._host)
    else:
        action = Action(ctx.CurrentQQ)

    if isinstance(ctx, GroupMsg):
        if voice_url:
            return action.sendGroupVoice(ctx.FromGroupId, voiceUrl=voice_url)
        elif voice_base64:
            return action.sendGroupVoice(ctx.FromGroupId, voiceBase64Buf=voice_base64)
        elif voice_path:
            return action.sendGroupVoice(
                ctx.FromGroupId, voiceBase64Buf=file_to_base64(voice_path)
            )
    if isinstance(ctx, FriendMsg):
        if voice_url:
            if ctx.TempUin:
                return action.sendPrivateVoice(
                    ctx.FromUin, ctx.TempUin, voiceUrl=voice_url
                )
            elif ctx.MsgType == MsgTypes.PhoneMsg:  # 来自手机的消息
                return None
            else:
                return action.sendFriendVoice(ctx.FromUin, voiceUrl=voice_url)
        elif voice_base64:
            if ctx.TempUin:
                return action.sendPrivateVoice(
                    ctx.FromUin, ctx.TempUin, voiceBase64Buf=voice_base64
                )
            elif ctx.MsgType == MsgTypes.PhoneMsg:  # 来自手机的消息
                return None
            else:
                return action.sendFriendVoice(ctx.FromUin, voiceBase64Buf=voice_base64)
        elif voice_path:
            if ctx.TempUin:
                return action.sendPrivateVoice(
                    ctx.FromUin, ctx.TempUin, voiceBase64Buf=file_to_base64(voice_path)
                )
            elif ctx.MsgType == MsgTypes.PhoneMsg:  # 来自手机的消息
                return None
            else:
                return action.sendFriendVoice(
                    ctx.FromUin, voiceBase64Buf=file_to_base64(voice_path)
                )
    return None



import importlib
import os
import re
from types import ModuleType
from typing import Dict, List

from prettytable import PrettyTable

from botoy import json


class Plugin:
    def __init__(self, module: ModuleType):
        self.module = module

    def reload(self):
        self.module = importlib.reload(self.module)

    @property
    def name(self):
        return self.module.__name__.split('.')[-1][4:]

    @property
    def receive_group_msg(self):
        return self.module.__dict__.get('receive_group_msg')

    @property
    def receive_friend_msg(self):
        return self.module.__dict__.get('receive_friend_msg')

    @property
    def receive_events(self):
        return self.module.__dict__.get('receive_events')


class PluginManager:
    def __init__(self, plugin_dir: str = 'plugins'):
        self.plugin_dir = plugin_dir
        self._plugins: Dict[str, Plugin] = dict()
        self._removed_plugins: Dict[str, Plugin] = dict()

        # 本地缓存的停用的插件名称列表
        self._load_removed_plugin_names()

    def _load_removed_plugin_names(self):
        """读取已移除插件名文件，初始化_removed_plugins属性，后面刷新插件时不再加载"""
        if os.path.exists('REMOVED_PLUGINS'):
            with open('REMOVED_PLUGINS', encoding='utf8') as f:
                self._removed_plugin_names = json.load(f)['plugins']
        else:
            with open('REMOVED_PLUGINS', 'w', encoding='utf8') as f:
                json.dump(
                    {'tips': '用于存储已停用插件信息,请不要修改这个文件', 'plugins': []},
                    f,
                    ensure_ascii=False,
                )
            self._removed_plugin_names = []

    def _update_removed_plugin_names(self):
        """更新已移除插件名文件"""
        data = {
            'tips': '用于存储已停用插件信息,请不要修改这个文件',
            'plugins': list(set(self._removed_plugin_names)),  # 去重，虽然显得多余
        }
        with open('REMOVED_PLUGINS', 'w', encoding='utf8') as f:
            json.dump(data, f, ensure_ascii=False)

    def load_plugins(self, plugin_dir: str = None) -> None:
        """加载插件，只会加载新插件"""
        if plugin_dir is None:
            plugin_dir = self.plugin_dir
        plugin_files = (
            i for i in os.listdir(plugin_dir) if re.search(r'^bot_\w+\.py$', i)
        )
        for plugin_file in plugin_files:
            module = importlib.import_module(
                '{}.{}'.format(plugin_dir.replace('/', '.'), plugin_file.split('.')[0])
            )
            plugin = Plugin(module)
            if plugin.name in self._removed_plugin_names:
                self._removed_plugins[plugin.name] = plugin
            else:
                self._plugins[plugin.name] = plugin

    def reload_plugins(self, plugin_dir: str = None) -> None:
        """加载新插件并刷新旧插件"""
        # reload old
        old_plugins = self._plugins.copy()
        for old_plugin in old_plugins.values():
            old_plugins[old_plugin.name].reload()
        # load new
        self.load_plugins(plugin_dir)
        # tidy
        self._plugins.update(old_plugins)

    def reload_plugin(self, plugin_name: str) -> None:
        """根据指定插件名刷新插件，不管是否存在，都不会报错"""
        if plugin_name in self._plugins:
            self._plugins[plugin_name].reload()

    def remove_plugin(self, plugin_name: str) -> None:
        """移除指定插件, 不会报错"""
        try:
            if plugin_name in self._plugins:
                self._removed_plugins[plugin_name] = self._plugins.pop(plugin_name)
                # 缓存到本地
                self._removed_plugin_names.append(plugin_name)
                self._update_removed_plugin_names()
        except KeyError:  # 可能由self._removed_plugins[plugin_name]引发
            pass

    def recover_plugin(self, plugin_name: str) -> None:
        """重新开启指定插件, 不会报错"""
        try:
            if plugin_name in self._removed_plugins:
                self._plugins[plugin_name] = self._removed_plugins.pop(plugin_name)
                if plugin_name in self._removed_plugin_names:
                    self._removed_plugin_names.remove(plugin_name)
                    self._update_removed_plugin_names()
        except KeyError:
            pass

    @property
    def plugins(self) -> List[str]:
        '''return a list of plugin name'''
        return list(self._plugins)

    @property
    def removed_plugins(self) -> List[str]:
        '''return a list of removed plugin name'''
        return list(self._removed_plugins)

    @property
    def friend_msg_receivers(self):
        '''funcs to handle (friend msg)context'''
        return [
            plugin.receive_friend_msg
            for plugin in self._plugins.values()
            if plugin.receive_friend_msg
        ]

    @property
    def group_msg_receivers(self):
        '''funcs to handle (group msg)context'''
        return [
            plugin.receive_group_msg
            for plugin in self._plugins.values()
            if plugin.receive_group_msg
        ]

    @property
    def event_receivers(self):
        '''funcs to handle (event msg)context'''
        return [
            plugin.receive_events
            for plugin in self._plugins.values()
            if plugin.receive_events
        ]

    @property
    def info_table(self) -> str:
        table = PrettyTable(['Receiver', 'Count', 'Info'])
        table.add_row(
            [
                'Friend Msg Receiver',
                len(self.friend_msg_receivers),
                '/'.join(
                    [
                        f'{p.name}'
                        for p in self._plugins.values()
                        if p.receive_friend_msg
                    ]
                ),
            ]
        )
        table.add_row(
            [
                'Group  Msg Receiver',
                len(self.group_msg_receivers),
                '/'.join(
                    [f'{p.name}' for p in self._plugins.values() if p.receive_group_msg]
                ),
            ]
        )
        table.add_row(
            [
                'Event      Receiver',
                len(self.event_receivers),
                '/'.join(
                    [f'{p.name}' for p in self._plugins.values() if p.receive_events]
                ),
            ]
        )
        table_removed = PrettyTable(['Removed Plugins'])
        table_removed.add_row(['/'.join(self.removed_plugins)])
        return str(table) + '\n' + str(table_removed)



from ..model import GroupMsg


def ignore_these_groups(*groups):
    """不接受这些群组的消息 GroupMsg"""

    def deco(func):
        def inner(ctx):
            nonlocal groups
            assert isinstance(ctx, GroupMsg)
            from_group = ctx.FromGroupId
            if from_group not in groups:
                return func(ctx)
            return None

        return inner

    return deco



# pylint: disable=C0415,C0413
__version__ = '0.0.23'


def check_version():
    def _check_version():
        import re
        from distutils.version import LooseVersion as V

        import httpx

        try:
            resp = httpx.get(
                'https://mirrors.aliyun.com/pypi/simple/botoy/', timeout=10
            )
            resp.raise_for_status()
        except Exception:
            pass
        else:
            versions = re.findall(r'botoy-(.*?)\.tar\.gz', resp.text)
            if versions:
                versions = set(versions)
            local_v = V(__version__)
            latest_version = max(V(v) for v in versions)
            if local_v < latest_version:
                info = f'\033[33m==== 当前版本为: \033[31m{local_v}\033[33m, 已有最新版本: \033[31m{latest_version}\033[33m, 请及时更新! ====\033[0m'
                print(info)

    from threading import Thread

    t = Thread(target=_check_version)
    t.setDaemon(True)
    t.start()



from ..model import FriendMsg, GroupMsg


def ignore_botself(func=None):
    """忽略机器人自身的消息 GroupMsg, FriendMsg"""
    if func is None:
        return ignore_botself

    def inner(ctx):
        assert isinstance(ctx, (GroupMsg, FriendMsg))
        if isinstance(ctx, GroupMsg):
            userid = ctx.FromUserId
        else:
            userid = ctx.FromUin
        if userid != ctx.CurrentQQ:
            return func(ctx)
        return None

    return inner



"""辅助构建发送消息宏"""
import random
from collections.abc import Sequence
from typing import List, Union


def atUser(users: Union[List[int], int]) -> str:
    """艾特(ATUSER)
    :param users: 需要艾特的QQ号，如需多个请传列表
    """
    if not isinstance(users, Sequence):
        users = [users]
    return '[ATUSER(%s)]' % ','.join([str(i) for i in users])


def picFlag() -> str:
    """改变图片顺序"""
    return '[PICFLAG]'


def showPic(code: int = None) -> str:
    """秀图宏
    :param code: 秀图样式编号 40000-40006, 如不传则随机选取一个
            40000 秀图  40001 幻影  40002 抖动
            40003 生日  40004 爱你  40005 征友
            40006 无(只显示大图无特效)
    """
    if code is None:
        code = random.randint(40000, 40007)
    return '[秀图%d]' % code


def getUserNick(user: int) -> str:
    """获取用户昵称"""
    return '[GETUSERNICK(%d)]' % user



class MsgTypes:
    """消息类型"""

    AtMsg = 'AtMsg'
    PicMsg = 'PicMsg'
    TextMsg = 'TextMsg'
    ReplyMsg = 'ReplyMsg'
    ReplyMsgA = 'ReplayMsg'
    VoiceMsg = 'VoiceMsg'
    VideoMsg = 'VideoMsg'
    PhoneMsg = 'PhoneMsg'
    RedBagMsg = 'RedBagMsg'
    TempSessionMsg = 'TempSessionMsg'
    FriendFileMsg = 'FriendFileMsg'
    GroupFileMsg = 'GroupFileMsg'


class EventNames:
    """事件类型"""

    ON_EVENT_GROUP_REVOKE = 'ON_EVENT_GROUP_REVOKE'
    ON_EVENT_GROUP_EXIT = 'ON_EVENT_GROUP_EXIT'
    ON_EVENT_GROUP_JOIN = 'ON_EVENT_GROUP_JOIN'
    ON_EVENT_GROUP_ADMINSYSNOTIFY = 'ON_EVENT_GROUP_ADMINSYSNOTIFY'  #  群系统消息相关 通知内容
    ON_EVENT_FRIEND_REVOKE = 'ON_EVENT_FRIEND_REVOKE'  # 好友撤回消息
    ON_EVENT_FRIEND_DELETE = 'ON_EVENT_FRIEND_DELETE'  # 删除好友
    ON_EVENT_GROUP_SHUT = 'ON_EVENT_GROUP_SHUT'  # 群禁言
    ON_EVENT_GROUP_ADMIN = 'ON_EVENT_GROUP_ADMIN'  # 群管变更事件
    ON_EVENT_NOTIFY_PUSHADDFRD = 'ON_EVENT_NOTIFY_PUSHADDFRD'  # 添加好友成功后的反馈
    ON_EVENT_FRIEND_ADDED = 'ON_EVENT_FRIEND_ADDED'  # 收到好友请求事件
    ON_EVENT_GROUP_EXIT_SUCC = 'ON_EVENT_GROUP_EXIT_SUCC'  # 主动退出群聊


# pylint: disable=non-ascii-name
class Emoticons:
    """表情"""

    k歌 = '[表情140]'
    爱你 = '[表情122]'
    爱情 = '[表情42]'
    爱心 = '[表情66]'
    傲慢 = '[表情23]'
    白眼 = '[表情22]'
    棒棒糖 = '[表情142]'
    抱拳 = '[表情118]'
    爆筋 = '[表情142]'
    鄙视 = '[表情105]'
    闭嘴 = '[表情7]'
    鞭炮 = '[表情137]'
    便便 = '[表情59]'
    不 = '[表情123]'
    搽汗 = '[表情97]'
    彩球 = '[表情164]'
    菜刀 = '[表情112]'
    差劲 = '[表情121]'
    钞票 = '[表情158]'
    车厢 = '[表情154]'
    呲牙 = '[表情13]'
    打伞 = '[表情163]'
    大兵 = '[表情29]'
    大哭 = '[表情9]'
    蛋糕 = '[表情53]'
    刀 = '[表情56]'
    得意 = '[表情4]'
    灯笼 = '[表情138]'
    灯泡 = '[表情160]'
    凋谢 = '[表情64]'
    多云 = '[表情156]'
    发财 = '[表情139]'
    发呆 = '[表情3]'
    发抖 = '[表情41]'
    发怒 = '[表情11]'
    饭 = '[表情61]'
    飞机 = '[表情151]'
    飞吻 = '[表情85]'
    奋斗 = '[表情30]'
    风车 = '[表情161]'
    尴尬 = '[表情10]'
    高铁右车头 = '[表情155]'
    高铁左车头 = '[表情153]'
    勾引 = '[表情119]'
    购物 = '[表情141]'
    鼓掌 = '[表情99]'
    哈欠 = '[表情104]'
    害羞 = '[表情6]'
    憨笑 = '[表情28]'
    好 = '[表情124]'
    喝彩 = '[表情144]'
    喝奶 = '[表情148]'
    坏笑 = '[表情101]'
    揉手 = '[表情129]'
    回头 = '[表情127]'
    饥饿 = '[表情24]'
    激动 = '[表情130]'
    街舞 = '[表情131]'
    惊恐 = '[表情26]'
    惊讶 = '[表情0]'
    咖啡 = '[表情60]'
    开车 = '[表情152]'
    磕头 = '[表情126]'
    可爱 = '[表情21]'
    可怜 = '[表情111]'
    抠鼻 = '[表情98]'
    骷髅 = '[表情37]'
    酷 = '[表情16]'
    快哭了 = '[表情107]'
    困 = '[表情25]'
    篮球 = '[表情114]'
    冷汗 = '[表情96]'
    礼物 = '[表情69]'
    流汗 = '[表情27]'
    流泪 = '[表情5]'
    玫瑰 = '[表情63]'
    难过 = '[表情15]'
    闹钟 = '[表情162]'
    怄火 = '[表情86]'
    啤酒 = '[表情113]'
    瓢虫 = '[表情117]'
    撇嘴 = '[表情1]'
    乒乓 = '[表情115]'
    祈祷 = '[表情145]'
    强 = '[表情76]'
    敲打 = '[表情38]'
    亲亲 = '[表情109]'
    青蛙 = '[表情170]'
    糗大了 = '[表情100]'
    拳头 = '[表情120]'
    弱 = '[表情77]'
    色 = '[表情2]'
    沙发 = '[表情166]'
    闪电 = '[表情54]'
    胜利 = '[表情79]'
    示爱 = '[表情110]'
    手枪 = '[表情169]'
    衰 = '[表情36]'
    帅 = '[表情143]'
    双喜 = '[表情136]'
    睡 = '[表情8]'
    太阳 = '[表情74]'
    调皮 = '[表情12]'
    跳绳 = '[表情128]'
    跳跳 = '[表情43]'
    偷笑 = '[表情20]'
    吐 = '[表情19]'
    微笑 = '[表情14]'
    委屈 = '[表情106]'
    握手 = '[表情78]'
    西瓜 = '[表情89]'
    下面 = '[表情149]'
    下雨 = '[表情157]'
    吓 = '[表情110]'
    献吻 = '[表情132]'
    香蕉 = '[表情150]'
    心碎 = '[表情67]'
    熊猫 = '[表情159]'
    药 = '[表情168]'
    疑问 = '[表情32]'
    阴险 = '[表情108]'
    拥抱 = '[表情49]'
    邮件 = '[表情142]'
    右哼哼 = '[表情103]'
    右太极 = '[表情134]'
    月亮 = '[表情75]'
    晕 = '[表情34]'
    再见 = '[表情39]'
    炸弹 = '[表情55]'
    折磨 = '[表情35]'
    纸巾 = '[表情167]'
    咒骂 = '[表情31]'
    猪头 = '[表情46]'
    抓狂 = '[表情18]'
    转圈 = '[表情125]'
    足球 = '[表情57]'
    钻戒 = '[表情165]'
    左哼哼 = '[表情102]'
    左太极 = '[表情133]'



# pylint: disable = too-many-instance-attributes, W0212
import copy
import functools
import sys
import traceback
from collections.abc import Sequence
from typing import Any, Callable, List, Optional, Tuple, Union

import socketio

from botoy.config import Config
from botoy.log import enble_log_file, logger
from botoy.model import EventMsg, FriendMsg, GroupMsg
from botoy.plugin import PluginManager
from botoy.pool import WorkerPool

#######################
#     socketio
#        | dict
#  message handler
#        | context
#  context middleware
#        | new context
#     receiver
#######################


class Botoy:
    """
    :param qq: 机器人QQ号(多Q就传qq号列表), 如果不传则会监听服务端传过来的所有机器人的
                所有信息，如果传了，则只会接收对应机器人的信息
    :param use_plugins: 是否开启插件功能, 默认``False``
    :param port: 运行端口, 默认为``8888``
    :param host: ip，默认为``http://127.0.0.1``
    :param group_blacklist: 群黑名单, 此名单中的群聊消息不会被处理,默认为``空``
    :param friend_blacklist: 好友黑名单，此名单中的好友消息不会被处理，默认为``空``
    :param blocked_users: 用户黑名单，即包括群消息和好友消息, 该用户的消息都不会处理, 默认为``空``
    :param log: 是否开启日志
    :param log_file: 是否输出日志文件
    """

    def __init__(
        self,
        *,
        qq: Union[int, List[int]] = None,
        use_plugins: bool = False,
        port: int = None,
        host: str = None,
        group_blacklist: List[int] = None,
        friend_blacklist: List[int] = None,
        blocked_users: List[int] = None,
        log: bool = True,
        log_file: bool = True,
    ):
        if qq is not None:
            if isinstance(qq, Sequence):
                self.qq = list(qq)
            else:
                self.qq = [qq]
        else:
            self.qq = None

        self.use_plugins = use_plugins
        self.config = Config(
            host, port, group_blacklist, friend_blacklist, blocked_users
        )

        # 作为程序是否应该退出的标志，以便后续用到
        self._exit = False

        if log:
            if log_file:
                enble_log_file()
        else:
            logger.disable(__name__)

        # 消息接收函数列表
        # 这里只储存主体文件中通过装饰器或函数添加的接收函数
        self._friend_msg_receivers = []
        self._group_msg_receivers = []
        self._event_receivers = []

        # 消息上下文对象中间件列表
        # 中间件以对应消息上下文为唯一参数，返回值与上下文类型一致则向下传递
        # 否则直接丢弃该次消息
        self._friend_context_middlewares = []
        self._group_context_middlewares = []
        self._event_context_middlewares = []

        # webhook
        if self.config.webhook:
            from . import webhook  # pylint: disable=C0415

            self._friend_msg_receivers.append(webhook.receive_friend_msg)
            self._group_msg_receivers.append(webhook.receive_group_msg)
            self._event_receivers.append(webhook.receive_events)

        # 插件管理
        # 管理插件提供的接收函数
        self.plugMgr = PluginManager()
        if use_plugins:
            self.plugMgr.load_plugins()
            print(self.plugin_status)

        # 当连接上或断开连接运行的函数
        # 如果通过装饰器注册了, 这两个字段设置成(func, every_time)
        # func 是需要执行的函数， every_time 表示是否每一次连接或断开都会执行
        self._when_connected_do: Tuple[Callable, bool] = None
        self._when_disconnected_do: Tuple[Callable, bool] = None

        # 线程池 TODO: 开放该参数
        thread_works = 50
        self.pool = WorkerPool(thread_works)

        # 依次各种初始化
        self._initialize_socketio()
        self._initialize_handlers()

    ########################################################################
    # Add context receivers
    ########################################################################
    def on_friend_msg(self, receiver: Callable[[FriendMsg], Any]):
        """添加好友消息接收函数"""
        self._friend_msg_receivers.append(receiver)
        return self  # 包括下面的六个方法是都不需要返回值的, 但返回本身也无妨,可以支持链式初始化

    def on_group_msg(self, receiver: Callable[[GroupMsg], Any]):
        """添加群消息接收函数"""
        self._group_msg_receivers.append(receiver)
        return self

    def on_event(self, receiver: Callable[[EventMsg], Any]):
        """添加事件消息接收函数"""
        self._event_receivers.append(receiver)
        return self

    ########################################################################
    # Add context middlewares
    ########################################################################
    def friend_context_use(
        self, middleware: Callable[[FriendMsg], Optional[FriendMsg]]
    ):
        """注册好友消息中间件"""
        self._friend_context_middlewares.append(middleware)
        return self

    def group_context_use(self, middleware: Callable[[GroupMsg], Optional[GroupMsg]]):
        """注册群消息中间件"""
        self._group_context_middlewares.append(middleware)
        return self

    def event_context_use(self, middleware: Callable[[EventMsg], Optional[EventMsg]]):
        """注册事件消息中间件"""
        self._event_context_middlewares.append(middleware)
        return self

    ########################################################################
    # shortcuts to call plugin manager methods
    ########################################################################
    # 只推荐使用这几个方法，其他的更细致的方法需要通过 plugMgr 对象访问
    # 原始plugMgr提供了自由度更高的方法
    def load_plugins(self):
        '''加载新插件'''
        return self.plugMgr.load_plugins()

    def reload_plugins(self):
        '''重载旧插件，加载新插件'''
        return self.plugMgr.reload_plugins()

    def reload_plugin(self, plugin_name: str):
        '''重载指定插件'''
        return self.plugMgr.reload_plugin(plugin_name)

    def remove_plugin(self, plugin_name: str):
        '''停用指定插件'''
        return self.plugMgr.remove_plugin(plugin_name)

    def recover_plugin(self, plugin_name: str):
        '''启用指定插件'''
        return self.plugMgr.recover_plugin(plugin_name)

    @property
    def plugin_status(self):
        '''插件启用状态'''
        return self.plugMgr.info_table

    @property
    def plugins(self):
        '''插件名列表'''
        return self.plugMgr.plugins

    @property
    def removed_plugins(self):
        '''已停用的插件名列表'''
        return self.plugMgr.removed_plugins

    ##########################################################################
    # decorators for registering hook function when connected or disconnected
    ##########################################################################
    def when_connected(self, func: Callable = None, *, every_time=False):
        if func is None:
            return functools.partial(self.when_connected, every_time=every_time)
        self._when_connected_do = (func, every_time)
        return None

    def when_disconnected(self, func: Callable = None, *, every_time=False):
        if func is None:
            return functools.partial(self.when_disconnected, every_time=every_time)
        self._when_disconnected_do = (func, every_time)
        return None

    ########################################################################
    # about socketio
    ########################################################################
    def connect(self):
        logger.success('Connected to the server successfully!')

        # 连接成功执行用户定义的函数，如果有
        if self._when_connected_do is not None:
            self._when_connected_do[0]()
            # 如果不需要每次运行，这里运行一次后就废弃设定的函数
            if not self._when_connected_do[1]:
                self._when_connected_do = None

    def disconnect(self):
        logger.warning('Disconnected to the server!')
        # 断开连接后执行用户定义的函数，如果有
        if self._when_disconnected_do is not None:
            self._when_disconnected_do[0]()
            if not self._when_disconnected_do[1]:
                self._when_disconnected_do = None

    def close(self, status=0):
        # 如果先关线程池，当新消息进入时依然会调用submit方法,此时会报错
        self.socketio.disconnect()
        self.pool.shutdown(wait=False)
        self._exit = True
        sys.exit(status)

    def run(self):
        logger.info('Connecting to the server...')
        try:
            self.socketio.connect(self.config.address, transports=['websocket'])
        except Exception:
            logger.error(traceback.format_exc())
            self.close(1)
        else:
            try:
                self.socketio.wait()
            except KeyboardInterrupt:
                pass
            finally:
                print('bye~')
                self.close(0)

    ########################################################################
    # context distributor
    ########################################################################
    def _friend_context_distributor(self, context: FriendMsg):
        for f_receiver in [
            *self._friend_msg_receivers,
            *self.plugMgr.friend_msg_receivers,
        ]:
            self.pool.submit(f_receiver, copy.deepcopy(context))

    def _group_context_distributor(self, context: GroupMsg):
        for g_receiver in [
            *self._group_msg_receivers,
            *self.plugMgr.group_msg_receivers,
        ]:
            self.pool.submit(g_receiver, copy.deepcopy(context))

    def _event_context_distributor(self, context: EventMsg):
        for e_receiver in [
            *self._event_receivers,
            *self.plugMgr.event_receivers,
        ]:
            self.pool.submit(e_receiver, copy.deepcopy(context))

    ########################################################################
    # message handler
    ########################################################################
    def _friend_msg_handler(self, msg):
        context = FriendMsg(msg)
        if self.qq and context.CurrentQQ not in self.qq:
            return
        logger.info(f'{context.__class__.__name__} ->  {context.data}')
        # 黑名单
        if context.FromUin in self.config.friend_blacklist:
            return
        # 屏蔽用户
        if context.FromUin in self.config.blocked_users:
            return
        # 中间件
        if self._friend_context_middlewares:
            for middleware in self._friend_context_middlewares:
                new_context = middleware(context)
                if isinstance(new_context, type(context)):
                    context = new_context
                else:
                    return
        # 传递几个数据供(插件中的)接收函数调用, 其他不再注释
        context._host = self.config.host
        context._port = self.config.port
        self.pool.submit(self._friend_context_distributor, context)

    def _group_msg_handler(self, msg):
        context = GroupMsg(msg)
        if self.qq and context.CurrentQQ not in self.qq:
            return
        logger.info(f'{context.__class__.__name__} ->  {context.data}')
        # 黑名单
        if context.FromGroupId in self.config.group_blacklist:
            return
        # 屏蔽用户
        if context.FromUserId in self.config.blocked_users:
            return
        # 中间件
        if self._group_context_middlewares:
            for middleware in self._group_context_middlewares:
                new_context = middleware(context)
                if isinstance(new_context, type(context)):
                    context = new_context
                else:
                    return
        context._host = self.config.host
        context._port = self.config.port
        self.pool.submit(self._group_context_distributor, context)

    def _event_handler(self, msg):
        context = EventMsg(msg)
        if self.qq and context.CurrentQQ not in self.qq:
            return
        logger.info(f'{context.__class__.__name__} ->  {context.data}')
        # 中间件
        if self._event_context_middlewares:
            for middleware in self._event_context_middlewares:
                new_context = middleware(context)
                if isinstance(new_context, type(context)):
                    context = new_context
                else:
                    return
        context._host = self.config.host
        context._port = self.config.port
        self.pool.submit(self._event_context_distributor, context)

    ########################################################################
    # initialize
    ########################################################################
    def _initialize_socketio(self):
        self.socketio = socketio.Client()
        self.socketio.event()(self.connect)
        self.socketio.event()(self.disconnect)

    def _initialize_handlers(self):
        self.socketio.on('OnGroupMsgs')(self._group_msg_handler)
        self.socketio.on('OnFriendMsgs')(self._friend_msg_handler)
        self.socketio.on('OnEvents')(self._event_handler)

    ########################################################################
    def __repr__(self):
        return 'Botoy <{}> <host-{}> <port-{}> <address-{}>'.format(
            " ".join([str(i) for i in self.qq]),
            self.config.host,
            self.config.port,
            self.config.address,
        )



from ..model import FriendMsg, GroupMsg
from ..refine import refine_pic_friend_msg, refine_pic_group_msg


def startswith(string: str):
    """Content以指定前缀开头  GroupMsg, FriendMsg
    :param string: 前缀字符串, 会解析图片消息的Content
    """

    def deco(func):
        def inner(ctx):
            assert isinstance(ctx, (GroupMsg, FriendMsg))
            if isinstance(ctx, GroupMsg):
                refine_ctx = refine_pic_group_msg(ctx)  # type: _PicGroupMsg
            else:
                refine_ctx = refine_pic_friend_msg(ctx)  # type:_PicFriendMsg
            if refine_ctx is not None:
                content = refine_ctx.Content
            else:
                content = ctx.Content
            # 这里的content按理永远不可能为None，但就是碰到了这种情况，startswith用得比较多
            # 所以先在这里增加一步判断
            if content is not None and content.startswith(string):
                return func(ctx)
            return None

        return inner

    return deco



# pylint: disable=R0902,W0231
from typing import Dict, List

from botoy import json
from botoy.collection import MsgTypes
from botoy.exceptions import InvalidContextError
from botoy.model import GroupMsg
from botoy.refine import _copy_ctx


class _GroupMsg(GroupMsg):
    def _carry_properties(self, ctx: GroupMsg):
        self.message = ctx.message
        self.CurrentQQ = ctx.CurrentQQ

        self.data = ctx.data

        self.FromGroupId: int = ctx.FromGroupId
        self.FromGroupName: str = ctx.FromGroupName
        self.FromUserId: int = ctx.FromUserId
        self.FromNickName: str = ctx.FromNickName
        self.Content: str = ctx.Content
        self.MsgType: str = ctx.MsgType
        self.MsgTime: int = ctx.MsgTime
        self.MsgSeq: int = ctx.MsgSeq
        self.MsgRandom: int = ctx.MsgRandom
        self.RedBaginfo: dict = ctx.RedBaginfo


class _VoiceGroupMsg(_GroupMsg):
    """群语音消息"""

    def __init__(self, ctx: GroupMsg):
        voice_data = json.loads(ctx.Content)
        self.VoiceUrl: str = voice_data['Url']
        self.Tips: str = voice_data['Tips']
        super()._carry_properties(ctx)


class _VideoGroupMsg(_GroupMsg):
    """群视频消息"""

    def __init__(self, ctx: GroupMsg):
        video_data = json.loads(ctx.Content)
        self.ForwordBuf: str = video_data['ForwordBuf']
        self.ForwordField: int = video_data['ForwordField']
        self.Tips: str = video_data['Tips']
        self.VideoMd5: str = video_data['VideoMd5']
        self.VideoSize: str = video_data['VideoSize']
        self.VideoUrl: str = video_data['VideoUrl']
        super()._carry_properties(ctx)


class _ReplyGroupMsg(_GroupMsg):
    """群回复消息"""

    def __init__(self, ctx: GroupMsg):
        reply_data: Dict = json.loads(ctx.Content)
        super()._carry_properties(ctx)
        self.Content: str = reply_data.get('Content', '')
        self.OriginMsgSeq: int = reply_data.get('MsgSeq', -1)
        self.ReplyContent: str = reply_data.get('ReplayContent', '')
        self.SrcContent: str = reply_data.get('SrcContent', '')
        self.Tips: str = reply_data.get('Tips', '')
        self.AtUserID: List = reply_data.get('UserID', [])


class _GroupPic:
    def __init__(self, pic: dict):
        '''[{"FileId":2161733733,"FileMd5":"","FileSize":449416,"ForwordBuf":"","ForwordField":8,"Url":""}'''
        self.FileId: int = pic.get('FileId')
        self.FileMd5: str = pic.get('FileMd5')
        self.FileSize: int = pic.get('FileSize')
        self.ForwordBuf: str = pic.get('ForwordBuf')
        self.ForwordField: int = pic.get('ForwordField')
        self.Url: str = pic.get('Url')


class _PicGroupMsg(_GroupMsg):
    """群图片/表情包消息"""

    def __init__(self, ctx: GroupMsg):
        judge = {'[群图片]': 'GroupPic', '[好友图片]': 'FriendPic'}
        pic_data = json.loads(ctx.Content)
        self.Tips: str = pic_data['Tips']
        if self.Tips != '[群消息-QQ闪照]':
            self.GroupPic: List[_GroupPic] = [
                _GroupPic(i) for i in pic_data[judge[self.Tips]]
            ]
        else:
            self.GroupPic: List[_GroupPic] = [_GroupPic(pic_data)]
        super()._carry_properties(ctx)
        self.Content: str = pic_data.get('Content') or ''


class _AtGroupMsg(_GroupMsg):
    def __init__(self, ctx: GroupMsg):
        at_data: Dict = json.loads(ctx.Content)
        super()._carry_properties(ctx)
        self.Content: str = at_data.get('Content', '')
        self.OriginMsgSeq: int = at_data.get('MsgSeq', -1)
        self.SrcContent: str = at_data.get('SrcContent', '')
        self.Tips: str = at_data.get('Tips', '')
        self.AtUserID: List = at_data.get('UserID', [])


class _GroupFileMsg(_GroupMsg):
    def __init__(self, ctx: GroupMsg):
        file_data: Dict = json.loads(ctx.Content)
        super()._carry_properties(ctx)
        self.FileID: str = file_data.get('FileID', '')
        self.FileName: str = file_data.get('FileName', '')
        self.FileSize: int = file_data.get('FileSize', -1)
        self.Tips: str = file_data.get('Tips', '')


class _RedBagGroupMsg(_GroupMsg):
    """群红包消息"""

    def __init__(self, ctx: GroupMsg):
        redbag_info = ctx.RedBaginfo
        self.RedBag_Authkey: str = redbag_info.get('Authkey')
        self.RedBag_Channel: int = redbag_info.get('Channel')
        self.RedBag_Des: str = redbag_info.get('Des')
        self.RedBag_FromType: int = redbag_info.get('FromType')
        self.RedBag_FromUin: int = redbag_info.get('FromUin')
        self.RedBag_Listid: str = redbag_info.get('Listid')
        self.RedBag_RedType: int = redbag_info.get('RedType')
        self.RedBag_StingIndex: str = redbag_info.get('StingIndex')
        self.RedBag_Tittle: str = redbag_info.get('Tittle')
        self.RedBag_Token_17_2: str = redbag_info.get('Token_17_2')
        self.RedBag_Token_17_3: str = redbag_info.get('Token_17_3')
        super()._carry_properties(ctx)


@_copy_ctx
def refine_voice_group_msg(ctx: GroupMsg) -> _VoiceGroupMsg:
    """群语音消息"""
    if not isinstance(ctx, GroupMsg):
        raise InvalidContextError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.VoiceMsg:
        return _VoiceGroupMsg(ctx)
    return None


@_copy_ctx
def refine_video_group_msg(ctx: GroupMsg) -> _VideoGroupMsg:
    """群视频消息"""
    if not isinstance(ctx, GroupMsg):
        raise InvalidContextError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.VideoMsg:
        return _VideoGroupMsg(ctx)
    return None


@_copy_ctx
def refine_pic_group_msg(ctx: GroupMsg) -> _PicGroupMsg:
    """群图片/表情包消息"""
    if not isinstance(ctx, GroupMsg):
        raise InvalidContextError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.PicMsg:
        return _PicGroupMsg(ctx)
    return None


@_copy_ctx
def refine_RedBag_group_msg(ctx: GroupMsg) -> _RedBagGroupMsg:
    """群红包消息"""
    if not isinstance(ctx, GroupMsg):
        raise InvalidContextError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.RedBagMsg:
        return _RedBagGroupMsg(ctx)
    return None


@_copy_ctx
def refine_reply_group_msg(ctx: GroupMsg) -> _ReplyGroupMsg:
    """群回复消息"""
    if not isinstance(ctx, GroupMsg):
        raise InvalidContextError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType in (
        MsgTypes.ReplyMsg,
        MsgTypes.ReplyMsgA,
    ):  # Workaround for naming errors
        return _ReplyGroupMsg(ctx)
    return None


@_copy_ctx
def refine_at_group_msg(ctx: GroupMsg) -> _AtGroupMsg:
    """群回复消息"""
    if not isinstance(ctx, GroupMsg):
        raise InvalidContextError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.AtMsg:
        return _AtGroupMsg(ctx)
    return None


@_copy_ctx
def refine_file_group_msg(ctx: GroupMsg) -> _GroupFileMsg:
    """群文件消息"""
    if not isinstance(ctx, GroupMsg):
        raise InvalidContextError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.GroupFileMsg:
        return _GroupFileMsg(ctx)
    return None



import functools
import time
from collections import defaultdict
from queue import Queue
from threading import Thread

from botoy.log import logger


class TaskThread(Thread):
    def __init__(self):
        super().__init__()
        self.tasks = Queue(maxsize=-1)
        self.setDaemon(True)
        self.start()

    def run(self):
        while True:
            try:
                self.tasks.get()()
            except Exception as e:
                logger.warning(f'queued_up装饰器: 队列任务出错{e}')
            time.sleep(0.6)

    def put_task(self, target, *args):
        task = functools.partial(target, *args)
        self.tasks.put(task)


taskThread_dict = defaultdict(TaskThread)


def queued_up(func=None, *, name='default'):
    """队列执行函数
    :param name: 指定队列分组, 不同的名称用不同的队列
    """
    if func is None:
        return functools.partial(queued_up, name=name)

    def inner(ctx):
        task_thread = taskThread_dict[repr(name)]
        task_thread.put_task(func, ctx)

    return inner



import io

from setuptools import find_packages, setup

meta = {}

with io.open('./botoy/__version__.py', encoding='utf-8') as f:
    exec(f.read(), meta)  # pylint: disable=W0122


def read_files(files):
    data = []
    for file in files:
        with io.open(file, encoding='utf-8') as f:
            data.append(f.read())
    return "\n".join(data)


setup(
    name="botoy",
    description="OPQBot/IOTQQ/IOTBot的Python开发助手",
    long_description=read_files(['README.md', 'CHANGELOG.md']),
    long_description_content_type='text/markdown',
    version=meta['__version__'],
    author="wongxy",
    author_email="xiyao.wong@foxmail.com",
    url="https://github.com/xiyaowong/botoy",
    license='MIT',
    keywords=['iotbot', 'iotqq', 'OPQ', 'OPQBot'],
    packages=find_packages(),
    install_requires=read_files(['requirements.txt']),
    entry_points='''
        [console_scripts]
        botoy=botoy.cli:cli
    ''',
    python_requires='>=3.6',
)



# pylint: disable = too-many-instance-attributes, W0212,W0236,E1133
import asyncio
import copy
import sys
import traceback
from typing import Callable

import socketio

from botoy.client import Botoy
from botoy.log import logger
from botoy.model import EventMsg, FriendMsg, GroupMsg


class AsyncBotoy(Botoy):
    def _initialize_socketio(self):
        self.socketio = socketio.AsyncClient()
        self.socketio.event()(self.connect)
        self.socketio.event()(self.disconnect)

    def asyncRun(self, func: Callable, *args):
        self.pool.submit(func, *args)

    async def close(self, status=0):
        await self.socketio.disconnect()
        self.pool.shutdown(wait=False)
        self._exit = True
        sys.exit(status)

    async def run(self):
        try:
            await self.socketio.connect(self.config.address, transports=['websocket'])
        except Exception:
            logger.error(traceback.format_exc())
            await self.close(1)
        else:
            try:
                await self.socketio.wait()
            finally:
                print('bye~')
                await self.close(0)

    ########################################################################
    # context distributor
    ########################################################################
    async def _friend_context_distributor(self, context: FriendMsg):
        for f_receiver in [
            *self._friend_msg_receivers,
            *self.plugMgr.friend_msg_receivers,
        ]:
            new_context = copy.deepcopy(context)
            if asyncio.iscoroutinefunction(f_receiver):
                await f_receiver(new_context)
            else:
                self.asyncRun(f_receiver, new_context)

    async def _group_context_distributor(self, context: GroupMsg):
        for g_receiver in [
            *self._group_msg_receivers,
            *self.plugMgr.group_msg_receivers,
        ]:
            new_context = copy.deepcopy(context)
            if asyncio.iscoroutinefunction(g_receiver):
                await g_receiver(new_context)
            else:
                self.asyncRun(g_receiver, new_context)

    async def _event_context_distributor(self, context: EventMsg):
        for e_receiver in [
            *self._event_receivers,
            *self.plugMgr.event_receivers,
        ]:
            new_context = copy.deepcopy(context)
            if asyncio.iscoroutinefunction(e_receiver):
                await e_receiver(new_context)
            else:
                self.asyncRun(e_receiver, new_context)

    ########################################################################
    # message handler
    ########################################################################
    async def _friend_msg_handler(self, msg):
        context = FriendMsg(msg)
        if self.qq and context.CurrentQQ not in self.qq:
            return
        logger.info(f'{context.__class__.__name__} ->  {context.data}')
        # 黑名单
        if context.FromUin in self.config.friend_blacklist:
            return
        # 屏蔽用户
        if context.FromUin in self.config.blocked_users:
            return
        # 中间件
        if self._friend_context_middlewares:
            for middleware in self._friend_context_middlewares:
                if asyncio.iscoroutinefunction(middleware):
                    new_context = await middleware(context)
                else:
                    new_context = middleware(context)
                if isinstance(new_context, type(context)):
                    context = new_context
                else:
                    return
        context._host = self.config.host
        context._port = self.config.port
        await self._friend_context_distributor(context)

    async def _group_msg_handler(self, msg):
        context = GroupMsg(msg)
        if self.qq and context.CurrentQQ not in self.qq:
            return
        logger.info(f'{context.__class__.__name__} ->  {context.data}')
        # 黑名单
        if context.FromGroupId in self.config.group_blacklist:
            return
        # 屏蔽用户
        if context.FromUserId in self.config.blocked_users:
            return
        # 中间件
        if self._group_context_middlewares:
            for middleware in self._group_context_middlewares:
                if asyncio.iscoroutinefunction(middleware):
                    new_context = await middleware(context)
                else:
                    new_context = middleware(context)
                if isinstance(new_context, type(context)):
                    context = new_context
                else:
                    return
        context._host = self.config.host
        context._port = self.config.port
        await self._group_context_distributor(context)

    async def _event_handler(self, msg):
        context = EventMsg(msg)
        if self.qq and context.CurrentQQ not in self.qq:
            return
        logger.info(f'{context.__class__.__name__} ->  {context.data}')
        # 中间件
        if self._event_context_middlewares:
            for middleware in self._event_context_middlewares:
                if asyncio.iscoroutinefunction(middleware):
                    new_context = await middleware(context)
                else:
                    new_context = middleware(context)
                if isinstance(new_context, type(context)):
                    context = new_context
                else:
                    return
        context._host = self.config.host
        context._port = self.config.port
        await self._event_context_distributor(context)

    ########################################################################
    def __repr__(self):
        return 'AsyncBotoy <{}> <host-{}> <port-{}> <address-{}>'.format(
            " ".join([str(i) for i in self.qq]),
            self.config.host,
            self.config.port,
            self.config.address,
        )



import httpx

from botoy import EventMsg, FriendMsg, GroupMsg
from botoy.config import Config
from botoy.log import logger

# 以内置插件方式实现简单的webhook，将原始数据上报至指定地址
# 因为webhook功能只能通过配置文件开启，所以直接新建Config, 读取配置文件即可
# TODO: resolve response

_config = Config()
_c = httpx.Client(base_url=_config.address, timeout=_config.webhook_timeout)


def receive_friend_msg(ctx: FriendMsg):
    try:
        _c.post(httpx.URL(), json=ctx.message)
    except Exception as e:
        logger.warning('Webhook请求中的错误: %s' % e)


def receive_group_msg(ctx: GroupMsg):
    try:
        _c.post(httpx.URL(), json=ctx.message)
    except Exception as e:
        logger.warning('Webhook请求中的错误: %s' % e)


def receive_events(ctx: EventMsg):
    try:
        _c.post(httpx.URL(), json=ctx.message)
    except Exception as e:
        logger.warning('Webhook请求中的错误: %s' % e)



from typing import List

from botoy import json
from botoy.exceptions import InvalidConfigError
from botoy.util import check_schema

# | host             | ip或域名                           | 127.0.0.1 |
# | port             | 端口，也可以不需要                 | 8888      |
# | group_blacklist  | 群消息黑名单                       | []        |
# | friend_blacklist | 好友消息黑名单                     | []        |
# | blocked_users    | 用户黑名单，即包括群消息和好友消息 | []        |
# | webhook          | 是否开启webhook功能                | false     |
# | webhook_post_url | webhook上报地址                    | None      |
# | webhook_timeout  | webhook等待响应的延时              | 20        |


# 写得很乱，总之就是优先指定的参数值，然后文件中的配置值，最后默认值
class Config:
    def __init__(
        self,
        host: str = None,
        port: int = None,
        group_blacklist: List[int] = None,
        friend_blacklist: List[int] = None,
        blocked_users: List[int] = None,
        webhook: bool = None,
        webhook_post_url: str = None,
        webhook_timeout: int = None,
    ):
        self.host = host
        self.port = port
        self.group_blacklist = group_blacklist
        self.friend_blacklist = friend_blacklist
        self.blocked_users = blocked_users
        self.webhook = webhook
        self.webhook_post_url = webhook_post_url
        self.webhook_timeout = webhook_timeout
        self.load_file()

    def load_file(self):
        _c = {}
        try:
            with open('botoy.json', encoding='utf-8') as f:
                _c = json.load(f)
        except FileNotFoundError:
            pass
        except Exception as e:
            raise InvalidConfigError('配置文件不规范') from e

        # host
        if self.host is None:
            host = _c.get('host')
            if host is None:
                self.host = 'http://127.0.0.1'
            else:
                self.host = check_schema(host)
        self.host = self.host.strip('/')  # 防止手动设置host，末尾是/符号，而报错

        # port
        if self.port is None:
            port = _c.get('port')
            if port is None:  # 没有配置用默认值
                self.port = 8888
            elif port == 0:  # 配置为0, 则不用端口, 比如单独一个域名
                self.port = 0
            else:
                self.port = int(port)

        # group_blacklist
        if self.group_blacklist is None:
            self.group_blacklist = _c.get('group_blacklist') or list()

        # friend_blacklist
        if self.friend_blacklist is None:
            self.friend_blacklist = _c.get('friend_blacklist') or list()

        # blocked_users
        if self.blocked_users is None:
            self.blocked_users = _c.get('blocked_users') or list()

        # webhook
        if self.webhook is None:
            self.webhook = bool(_c.get('webhook'))

        if self.webhook:
            # webhook_post_url
            if self.webhook_post_url is None:
                webhook_post_url = _c.get('webhook_post_url')
                self.webhook_post_url = check_schema(webhook_post_url)
            else:
                self.webhook_post_url = check_schema(self.webhook_post_url)
            # webhook_timeout
            webhook_timeout = _c.get('webhook_timeout')
            if webhook_timeout is None:
                self.webhook_timeout = 20

    @property
    def address(self) -> str:
        # address 由host和port拼接
        if self.port == 0:
            return self.host
        return f'{self.host}:{self.port}'



"""一些用于接收函数的装饰器"""
from ._ensure_tempMsg import ensure_tempMsg
from ._equal_content import equal_content
from ._from_admin import from_admin
from ._from_botself import from_botself
from ._from_phone import from_phone
from ._from_these_groups import from_these_groups
from ._from_these_users import from_these_users
from ._ignore_botself import ignore_botself
from ._ignore_tempMsg import ignore_tempMsg
from ._ignore_these_groups import ignore_these_groups
from ._ignore_these_users import ignore_these_users
from ._in_content import in_content
from ._queued_up import queued_up
from ._startswith import startswith
from ._these_msgtypes import these_msgtypes
from ._with_pattern import with_pattern



import asyncio
import importlib
import json
import os
import pathlib
import sys
import textwrap

import click

from .__version__ import __version__
from .async_client import AsyncBotoy
from .client import Botoy
from .util import check_schema

echo = click.echo


@click.group()
def cli():
    """botoy 脚手架

    使用示例:

    1. 生成主体文件
    $ botoy init

    生成主体文件, 文件名为bot, qq号为123456:
    $ botoy init -n bot -q 123456

    2. 生成插件
    $ botoy add

    生成插件，名为test
    $ botoy add -n test

    生成插件，名为test, 只需要处理好友消息
    $ botoy add -n test -f

    生成插件，名为test, 需要处理好友消息, 群消息
    $ botoy add -n test -f -g

    可选项为-f -g -e 分别代表是否包含处理好友，群，事件，如果都不传，表示都要！

    3. 启动机器人(要求入口文件名为bot.py)
    $ botoy run
    """


@cli.command()
def version():
    """版本号"""
    echo(__version__)


@cli.command()
@click.option('-n', '--name', prompt="程序入口文件名", default='bot', show_default=True)
@click.option('-q', '--qq', prompt="机器人qq号", type=int, required=True)
@click.option(
    '--host', prompt='机器人运行host', default='http://127.0.0.1', show_default=True
)
@click.option('--port', prompt='机器人运行端口', default=8888, show_default=True, type=int)
def init(name, qq, host, port):
    """创建程序入口文件和配置文件"""
    plug = click.confirm('是否使用插件功能', default=True, show_default=True)
    template = textwrap.dedent(
        """
        from botoy import Action, Botoy, EventMsg, FriendMsg, GroupMsg

        qq = {qq}
        bot = Botoy(qq=qq, use_plugins={use_plugins})
        action = Action(qq)


        @bot.on_friend_msg
        def friend(ctx: FriendMsg):
            if ctx.Content == 'test':
                action.sendFriendText(ctx.FromUin, 'ok')


        @bot.on_group_msg
        def group(ctx: GroupMsg):
            if ctx.Content == 'test':
                action.sendGroupText(ctx.FromGroupId, 'ok')


        @bot.on_event
        def event(ctx: EventMsg):
            pass


        if __name__ == "__main__":
            bot.run()
    """
    ).format(qq=qq, use_plugins=plug)

    # main
    confirm = click.confirm(
        f'将生成 程序入口文件{name}.py 和 配置文件botoy.json, 这是覆盖写操作，是否继续?',
        default=False,
        show_default=True,
    )
    if not confirm:
        echo('操作已取消')
        sys.exit()

    with open(f'{name}.py', 'w', encoding='utf8') as f:
        f.write(template)
    echo(f'已生成{name}.py')
    echo(
        f"""
    运行如下命令:
        python {name}.py
    然后给机器人或机器人所在的群发送 test 不出意外，机器人会回复 ok!
    """
    )
    # config file
    config = {
        "host": check_schema(host),
        "port": port,
        "group_blacklist": [],
        "friend_blacklist": [],
        "blocked_users": [],
        "webhook": False,
        "webhook_post_url": "http://127.0.0.1:5000",
        "webhook_timeout": 20,
    }
    with open('botoy.json', 'w', encoding='utf8') as f:
        json.dump(config, f, indent='  ')

    if plug:
        if not os.path.isdir('plugins'):
            os.makedirs('plugins')


@cli.command()
@click.option('-n', '--name', prompt='插件名', required=True)
@click.option('-f', '--friend', is_flag=True, help='是否要接收好友消息?')
@click.option('-g', '--group', is_flag=True, help='是否要接收群消息?')
@click.option('-e', '--event', is_flag=True, help='是否要接收事件消息?')
def add(name, friend, group, event):
    """创建插件"""
    here = pathlib.Path('.').absolute()
    if here.name != 'plugins':
        if not os.path.isdir('plugins'):
            sys.exit('插件目录plugins不存在')
        plugin_dir = here / 'plugins'
    else:
        plugin_dir = here

    plugin_path = plugin_dir / f'bot_{name}.py'
    if os.path.exists(plugin_path):
        sys.exit('该插件已使用，请换一个插件名')
    if not any([friend, group, event]):
        friend = group = event = True

    imports = ['Action']
    receivers = []
    if friend:
        imports.append('FriendMsg')
        receivers.append(
            textwrap.dedent(
                """
                def receive_friend_msg(ctx: FriendMsg):
                    Action(ctx.CurrentQQ)
                """
            )
        )
    if group:
        imports.append('GroupMsg')
        receivers.append(
            textwrap.dedent(
                """
                def receive_group_msg(ctx: GroupMsg):
                    Action(ctx.CurrentQQ)
                """
            )
        )
    if event:
        imports.append('EventMsg')
        receivers.append(
            textwrap.dedent(
                """
                def receive_events(ctx: EventMsg):
                    Action(ctx.CurrentQQ)
                """
            )
        )
    with open(plugin_path, 'w', encoding='utf8') as f:
        f.write('from botoy import {}'.format(', '.join(imports)))
        f.write('\n\n')
        f.write('\n'.join(receivers))
    echo('ok')


@cli.command()
def run():
    """启动机器人
    要求入口文件名为bot.py
    """
    # look for bot.py
    if not pathlib.Path('bot.py').exists():
        sys.exit('该命令只接受入口文件命名为bot.py')
    # look for Botoy client
    sys.path.append(str(pathlib.Path('.').absolute()))
    module = importlib.import_module('bot')
    client = None
    for item in module.__dict__.values():
        if isinstance(item, (Botoy, AsyncBotoy)):
            client = item
            break
    else:
        sys.exit('无法找到(Async)Botoy对象')
    if client is not None:
        # 先判断子类
        if isinstance(client, AsyncBotoy):
            asyncio.run(client.run())
        else:
            client.run()


if __name__ == "__main__":
    cli()
